import ast
from copy import copy
from enum import IntFlag, auto
import itertools
from typing import Any, Callable, List, Literal, Optional, Tuple, Union

from scenic.core.errors import ScenicParseError, getText
import scenic.syntax.ast as s

# exposed functions


def compileScenicAST(
    scenicAST: ast.AST,
    *,
    filename: str = "<unknown>",
    inBehavior: bool = False,
    inMonitor: bool = False,
    inCompose: bool = False,
    inSetup: bool = False,
    inInterruptBlock: bool = False,
) -> Tuple[Union[ast.AST, List[ast.AST]], List[ast.AST]]:
    """Compiles Scenic AST to Python AST"""
    compiler = ScenicToPythonTransformer(filename)

    # set optional flags
    compiler.inBehavior = inBehavior
    compiler.inMonitor = inMonitor
    compiler.inCompose = inCompose
    compiler.inSetup = inSetup
    compiler.inInterruptBlock = inInterruptBlock

    tree = compiler.visit(scenicAST)
    if isinstance(tree, list):
        node = [ast.fix_missing_locations(n) for n in tree]
    else:
        node = ast.fix_missing_locations(tree)
    return node, compiler.requirements


# constants

temporaryName = "_Scenic_temporary_name"
behaviorArgName = "_Scenic_current_behavior"
checkPreconditionsName = "checkPreconditions"
checkInvariantsName = "checkInvariants"
interruptPrefix = "_Scenic_interrupt"

abortFlag = ast.Attribute(ast.Name("BlockConclusion", ast.Load()), "ABORT", ast.Load())
breakFlag = ast.Attribute(ast.Name("BlockConclusion", ast.Load()), "BREAK", ast.Load())
continueFlag = ast.Attribute(
    ast.Name("BlockConclusion", ast.Load()), "CONTINUE", ast.Load()
)
returnFlag = ast.Attribute(ast.Name("BlockConclusion", ast.Load()), "RETURN", ast.Load())
finishedFlag = ast.Attribute(
    ast.Name("BlockConclusion", ast.Load()), "FINISHED", ast.Load()
)

trackedNames = {"ego", "workspace"}
globalParametersName = "globalParameters"
builtinNames = {globalParametersName}


# shorthands for convenience

loadCtx = ast.Load()
ego = ast.Name("ego")

noArgs = ast.arguments(
    posonlyargs=[],
    args=[],
    vararg=None,
    kwonlyargs=[],
    kw_defaults=[],
    kwarg=None,
    defaults=[],
)
selfArg = ast.arguments(
    posonlyargs=[],
    args=[ast.arg(arg="self", annotation=None)],
    vararg=None,
    kwonlyargs=[],
    kw_defaults=[],
    kwarg=None,
    defaults=[],
)
initialBehaviorArgs = [
    ast.arg(arg=behaviorArgName, annotation=None),
    ast.arg(arg="self", annotation=None),
]
onlyBehaviorArgs = ast.arguments(
    posonlyargs=[],
    args=initialBehaviorArgs,
    vararg=None,
    kwonlyargs=[],
    kw_defaults=[],
    kwarg=None,
    defaults=[],
)

# helpers


class AttributeFinder(ast.NodeVisitor):
    """Utility class for finding all referenced attributes of a given name."""

    @staticmethod
    def find(target, node):
        af = AttributeFinder(target)
        af.visit(node)
        return af.attributes, af.rawLoc

    def __init__(self, target):
        super().__init__()
        self.target = target
        self.attributes = set()
        self.rawLoc = None

    def visit_Attribute(self, node):
        val = node.value
        if isinstance(val, ast.Name) and val.id == self.target:
            self.attributes.add(node.attr)
            return
        self.visit(val)

    def visit_Name(self, node):
        if node.id == self.target:
            self.rawLoc = node


class LocalFinder(ast.NodeVisitor):
    """Utility class for finding all local variables of a code block."""

    @staticmethod
    def findIn(block):
        lf = LocalFinder()
        for statement in block:
            lf.visit(statement)
        names = lf.names
        return names - lf.globals - lf.nonlocals

    def __init__(self):
        self.names = set()
        self.globals = set()
        self.nonlocals = set()

    def visit_Global(self, node):
        self.globals.update(node.names)

    def visit_Nonlocal(self, node):
        self.nonlocals.update(node.names)

    def visit_FunctionDef(self, node):
        self.names.add(node.name)
        self.visit(node.args)
        for decorator in node.decorator_list:
            self.visit(decorator)
        if node.returns is not None:
            self.visit(node.returns)
        # do not visit body; it's another block

    def visit_Lambda(self, node):
        self.visit(node.args)
        # do not visit body; it's another block

    def visit_ClassDef(self, node):
        self.names.add(node.name)
        for child in itertools.chain(node.bases, node.keywords, node.decorator_list):
            self.visit(child)
        # do not visit body; it's another block

    def visit_Import(self, node):
        for alias in node.names:
            bound = alias.asname
            if bound is None:
                bound = alias.name
            self.names.add(bound)

    def visit_ImportFrom(self, node):
        self.visit_Import(node)

    def visit_Name(self, node):
        if isinstance(node.ctx, (ast.Store, ast.Del)):
            self.names.add(node.id)

    def visit_ExceptHandler(self, node):
        if node.name is not None:
            self.names.add(node.name)
        self.generic_visit(node)


class Transformer(ast.NodeTransformer):
    """Subclass of `ast.NodeTransformer` with a method for raising syntax errors."""

    def __init__(self, filename):
        super().__init__()
        self.filename = filename

    def makeSyntaxError(self, msg, node: ast.AST) -> ScenicParseError:
        e = SyntaxError(msg)
        e.lineno = node.lineno
        e.offset = node.col_offset
        if node.end_lineno is not None:
            e.end_lineno = node.end_lineno
        if node.end_col_offset is not None:
            e.end_offset = node.end_col_offset
        e.filename = self.filename
        raise ScenicParseError(e)


# Proposition Constructors with Temporal Operators Support
PROPOSITION_AND = "PropositionAnd"
PROPOSITION_OR = "PropositionOr"
PROPOSITION_NOT = "PropositionNot"
ATOMIC_PROPOSITION = "AtomicProposition"
ALWAYS = "Always"
EVENTUALLY = "Eventually"
NEXT = "Next"
UNTIL = "Until"
IMPLIES = "Implies"
PROPOSITION_FACTORY = (
    PROPOSITION_AND,
    PROPOSITION_OR,
    PROPOSITION_NOT,
    ATOMIC_PROPOSITION,
    ALWAYS,
    EVENTUALLY,
    NEXT,
    UNTIL,
    IMPLIES,
)
TEMPORAL_PREFIX_OPS = {
    "always",
    "eventually",
    "next",
}


class PropositionTransformer(Transformer):
    def __init__(self, filename="<unknown>") -> None:
        super().__init__(filename)
        self.nextSyntaxId = 0

    def transform(
        self, node: ast.AST, nextSyntaxId=0
    ) -> Tuple[ast.AST, List[ast.AST], int]:
        """`transform` takes an AST node and apply transformations needed for temporal evaluation

        Args:
            node (ast.AST): AST node to perform proposition transformation
            nextSyntaxId (int, optional): Assign syntax ids starting at this number. Defaults to 0.

        Returns:
            ast.AST: Transformed AST node
        """
        self.nextSyntaxId = nextSyntaxId
        wrapped = self.visit(ast.fix_missing_locations(node))
        if self.is_proposition_factory(wrapped):
            return wrapped, self.nextSyntaxId
        newNode = self._create_atomic_proposition_factory(node)
        return newNode, self.nextSyntaxId

    def _register_requirement_syntax(self, syntax):
        """register requirement syntax for later use
        returns an ID for retrieving the syntax

        Args:
            propositionSyntax (ast.Node): AST Node that represents the requirement

        Returns:
            int: generated requirement syntax ID
        """
        syntaxId = self.nextSyntaxId
        self.nextSyntaxId += 1
        return syntaxId

    def _create_atomic_proposition_factory(self, node):
        """
        Given an expression, create an atomic proposition factory.

        Note: You must call `self.visit(node)` manually. This method does not make the surgeon visit the node.
        """
        lineNum = ast.Constant(node.lineno)
        ast.copy_location(lineNum, node)

        closure = ast.Lambda(noArgs, node)
        ast.copy_location(closure, node)

        syntaxId = self._register_requirement_syntax(node)
        syntaxIdConst = ast.Constant(syntaxId)
        ast.copy_location(syntaxIdConst, node)

        ap = ast.Call(
            func=ast.Name(id=ATOMIC_PROPOSITION, ctx=loadCtx),
            args=[closure],
            keywords=[
                ast.keyword(arg="syntaxId", value=syntaxIdConst),
            ],
        )
        ast.copy_location(ap, node)
        return ap

    def is_proposition_factory(self, node):
        return (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id in PROPOSITION_FACTORY
        )

    def visit_BoolOp(self, node: ast.BoolOp) -> ast.AST:
        """Convert a BoolOp node (`and`, `or`) to a corresponding proposition factory"""
        # 1. wrap each operand with a lambda function
        operands = []
        for operand in node.values:
            o = self.visit(operand)
            if self.is_proposition_factory(o):
                # if the operand is already an temporal requirement factory, keep it
                operands.append(self.visit(o))
                continue
            # if the operand is not an temporal requirement factory, make it an AP
            closure = self._create_atomic_proposition_factory(o)
            operands.append(closure)

        # 2. create a function call and pass operands
        boolOpToFunctionName = {
            ast.Or: "PropositionOr",
            ast.And: "PropositionAnd",
        }
        funcId = boolOpToFunctionName.get(type(node.op))
        newNode = ast.Call(
            func=ast.Name(id=funcId, ctx=ast.Load()),
            # pass a list of operands as the first argument
            args=[ast.copy_location(ast.List(elts=operands, ctx=ast.Load()), node)],
            keywords=[],
        )
        return ast.copy_location(newNode, node)

    def visit_UnaryOp(self, node):
        # rewrite `not` in requirements into a proposition factory
        if not isinstance(node.op, ast.Not):
            return self.generic_visit(node)

        lineNum = ast.Constant(node.lineno)
        ast.copy_location(lineNum, node)

        operand = self.visit(node.operand)

        newOperand = (
            operand
            if self.is_proposition_factory(operand)
            else self._create_atomic_proposition_factory(operand)
        )

        newNode = ast.Call(
            func=ast.Name(id=PROPOSITION_NOT, ctx=ast.Load()),
            args=[newOperand],
            keywords=[],
        )
        return ast.copy_location(newNode, node)

    def visit_Call(self, node: ast.Call):
        func = node.func
        if isinstance(func, ast.Name) and func.id in TEMPORAL_PREFIX_OPS:
            self.makeSyntaxError(
                f'malformed use of the "{func.id}" temporal operator', node
            )
        return self.generic_visit(node)

    def visit_Always(self, node: s.Always):
        value = self.visit(node.value)
        if not self.is_proposition_factory(value):
            value = self._create_atomic_proposition_factory(value)
        return ast.Call(
            func=ast.Name("Always", ctx=loadCtx),
            args=[value],
            keywords=[],
        )

    def visit_Eventually(self, node: s.Eventually):
        value = self.visit(node.value)
        if not self.is_proposition_factory(value):
            value = self._create_atomic_proposition_factory(value)
        return ast.Call(
            func=ast.Name("Eventually", ctx=loadCtx),
            args=[value],
            keywords=[],
        )

    def visit_Next(self, node: s.Next):
        value = self.visit(node.value)
        if not self.is_proposition_factory(value):
            value = self._create_atomic_proposition_factory(value)
        return ast.Call(
            func=ast.Name("Next", ctx=loadCtx),
            args=[value],
            keywords=[],
        )

    def visit_UntilOp(self, node: s.UntilOp):
        left = self.visit(node.left)
        if not self.is_proposition_factory(left):
            left = self._create_atomic_proposition_factory(left)
        right = self.visit(node.right)
        if not self.is_proposition_factory(right):
            right = self._create_atomic_proposition_factory(right)
        return ast.Call(
            func=ast.Name(id="Until", ctx=loadCtx),
            args=[self.visit(left), self.visit(right)],
            keywords=[],
        )

    def visit_ImpliesOp(self, node: s.ImpliesOp):
        hypothesis = self.visit(node.hypothesis)
        if not self.is_proposition_factory(hypothesis):
            hypothesis = self._create_atomic_proposition_factory(hypothesis)
        conclusion = self.visit(node.conclusion)
        if not self.is_proposition_factory(conclusion):
            conclusion = self._create_atomic_proposition_factory(conclusion)
        return ast.Call(
            func=ast.Name(id="Implies", ctx=loadCtx),
            args=[self.visit(hypothesis), self.visit(conclusion)],
            keywords=[],
        )


def unquote(s: str) -> str:
    if (s[:3] == s[-3:]) and s.startswith(("'''", '"""')):
        return s[3:-3]
    if (s[0] == s[-1]) and s.startswith(("'", '"')):
        return s[1:-1]
    return s


# transformer


class Context(IntFlag):
    TOP_LEVEL = auto()
    BEHAVIOR = auto()
    MONITOR = auto()
    COMPOSE = auto()
    DYNAMIC = BEHAVIOR | MONITOR | COMPOSE


class ScenicToPythonTransformer(Transformer):
    def __init__(self, filename) -> None:
        super().__init__(filename)

        self.requirements = []
        self.nextSyntaxId = 0

        self.inBehavior: bool = False
        "True if the transformer is processing behavior body"

        self.inMonitor: bool = False
        "True if the transformer is processing monitor body"

        self.behaviorLocals: set = set()
        "Set of variable names on the local scope of the behavior"

        self.inCompose: bool = False
        "True if the transformer is processing a `compose` block of modular scenario"

        self.inTryInterrupt = False
        self.inInterruptBlock = False
        self.inLoop = False
        self.usedBreak = False
        self.usedContinue = False

    @property
    def topLevel(self):
        return (
            not self.inBehavior
            and not self.inMonitor
            and not self.inCompose
            and not self.inTryInterrupt
        )

    def context(
        allowedContext: Context,
        errorBuilder: Optional[Callable[[str], str]] = None,
    ):
        "Mark AST node as only available inside certain contexts"

        def decorator(visitor: Callable[["ScenicToPythonTransformer", ast.AST], ast.AST]):
            def check_and_visit(self: "ScenicToPythonTransformer", node: ast.AST):
                ctx = None
                if self.topLevel and Context.TOP_LEVEL not in allowedContext:
                    ctx = "at the top level"
                elif self.inBehavior and Context.BEHAVIOR not in allowedContext:
                    ctx = "inside a behavior"
                elif self.inMonitor and Context.MONITOR not in allowedContext:
                    ctx = "inside a monitor"
                elif self.inCompose and Context.COMPOSE not in allowedContext:
                    ctx = "inside a compose block"
                if ctx:
                    raise self.makeSyntaxError(
                        (
                            f'Cannot use "{node.__class__.__name__}" {ctx}'
                            if errorBuilder is None
                            else errorBuilder(ctx)
                        ),
                        node,
                    )
                return visitor(self, node)

            return check_and_visit

        return decorator

    def generic_visit(self, node):
        if isinstance(node, s.AST):
            assert (
                False
            ), f'Scenic AST node "{node.__class__.__name__}" needs visitor in compiler'
        return super().generic_visit(node)

    # add support for list of nodes
    def visit(self, node: Union[ast.AST, List[ast.AST]]):
        if isinstance(node, ast.AST):
            return super().visit(node)
        elif isinstance(node, list):
            newStatements = []
            for statement in node:
                newStatement = self.visit(statement)
                if isinstance(newStatement, ast.AST):
                    newStatements.append(newStatement)
                else:
                    newStatements.extend(newStatement)
            return newStatements
        else:
            assert False, f"unknown object {node} encountered during compilation"

    # helper functions
    def _register_requirement_syntax(self, syntax: ast.AST) -> int:
        self.requirements.append(syntax)
        return len(self.requirements) - 1

    def visit_Name(self, node: ast.Name) -> Any:
        if node.id in builtinNames:
            if not isinstance(node.ctx, ast.Load):
                raise self.makeSyntaxError(f'unexpected keyword "{node.id}"', node)
            node = ast.copy_location(ast.Call(ast.Name(node.id, loadCtx), [], []), node)
        elif node.id in trackedNames:
            if not isinstance(node.ctx, ast.Load):
                raise self.makeSyntaxError(
                    f'only simple assignments to "{node.id}" are allowed', node
                )
            node = ast.copy_location(ast.Call(ast.Name(node.id, loadCtx), [], []), node)
        elif node.id in self.behaviorLocals:
            lookup = ast.Attribute(ast.Name(behaviorArgName, loadCtx), node.id, node.ctx)
            return ast.copy_location(lookup, node)
        return node

    # hook control flow nodes to set appropriate flags for try-interrupt

    def visit_For(self, node):
        old = self.inLoop
        self.inLoop = True
        newNode = self.generic_visit(node)
        self.inLoop = old
        return newNode

    def visit_While(self, node):
        old = self.inLoop
        self.inLoop = True
        newNode = self.generic_visit(node)
        self.inLoop = old
        return newNode

    def visit_FunctionDef(self, node):
        oldInLoop, oldInInterruptBlock = self.inLoop, self.inInterruptBlock
        self.inLoop, self.inInterruptBlock = False, False
        newNode = self.generic_visit(node)
        self.inLoop, self.inInterruptBlock = oldInLoop, oldInInterruptBlock
        return newNode

    def visit_Break(self, node):
        if self.inInterruptBlock and not self.inLoop:
            if not self.usedBreak:
                self.usedBreak = node
            newNode = ast.Return(breakFlag)
            return ast.copy_location(newNode, node)
        else:
            return self.generic_visit(node)

    def visit_Continue(self, node):
        if self.inInterruptBlock and not self.inLoop:
            if not self.usedContinue:
                self.usedContinue = node
            newNode = ast.Return(continueFlag)
            return ast.copy_location(newNode, node)
        else:
            return self.generic_visit(node)

    def visit_Return(self, node):
        if self.inInterruptBlock:
            value = ast.Constant(None) if node.value is None else node.value
            ret = ast.Return(ast.Call(returnFlag, [value], []))
            return ast.copy_location(ret, node)
        else:
            return self.generic_visit(node)

    def visit_Yield(self, node):
        if self.inCompose or self.inBehavior:
            # `yield` statements are not allowed inside a compose/behavior block
            raise self.makeSyntaxError(
                "Cannot use `yield` inside a compose/behavior block", node
            )
        return self.generic_visit(node)

    def visit_YieldFrom(self, node):
        if self.inCompose or self.inBehavior:
            # `yield from` statements are not allowed inside a compose/behavior block
            raise self.makeSyntaxError(
                "Cannot use `yield from` inside a compose/behavior block", node
            )
        return self.generic_visit(node)

    # Special Case

    @context(Context.DYNAMIC)
    def visit_TryInterrupt(self, node: s.TryInterrupt):
        statements = []
        oldInTryInterrupt = self.inTryInterrupt

        self.inTryInterrupt = True

        oldInInterruptBlock, oldInLoop = self.inInterruptBlock, self.inLoop
        self.inInterruptBlock = True
        self.inLoop = False
        self.usedBreak = False
        self.usedContinue = False

        def makeInterruptBlock(name, body):
            newBody = self.visit(body)
            allLocals = sorted(LocalFinder.findIn(newBody))  # Sort for determinism
            if allLocals:
                newBody.insert(0, ast.Nonlocal(allLocals))
            newBody.append(ast.Return(finishedFlag))
            return ast.FunctionDef(name, onlyBehaviorArgs, newBody, [], None)

        bodyName = f"{interruptPrefix}_body"
        statements.append(makeInterruptBlock(bodyName, node.body))

        handlerNames, conditionNames = [], []
        for i, handler in enumerate(node.interrupt_when_handlers):
            condition = handler.cond
            block = handler.body

            handlerName = f"{interruptPrefix}_handler_{i}"
            handlerNames.append(handlerName)
            conditionName = f"{interruptPrefix}_condition_{i}"
            conditionNames.append(conditionName)

            statements.append(makeInterruptBlock(handlerName, block))
            self.inGuard = True
            checker = ast.Lambda(noArgs, self.visit(condition))
            self.inGuard = False
            defChecker = ast.Assign([ast.Name(conditionName, ast.Store())], checker)
            statements.append(defChecker)

        self.inInterruptBlock, self.inLoop = oldInInterruptBlock, oldInLoop
        self.inTryInterrupt = oldInTryInterrupt

        # Prepare tuples of interrupt conditions and handlers
        # (in order from high priority to low, so reversed relative to the syntax)
        conditions = ast.Tuple(
            [ast.Name(n, ast.Load()) for n in reversed(conditionNames)], ast.Load()
        )
        handlers = ast.Tuple(
            [ast.Name(n, ast.Load()) for n in reversed(handlerNames)], ast.Load()
        )

        # Construct code to execute the try-interrupt statement
        args = [
            ast.Name(behaviorArgName, ast.Load()),
            ast.Name("self", ast.Load()),
            ast.Name(bodyName, ast.Load()),
            conditions,
            handlers,
        ]
        callRuntime = ast.Call(ast.Name("runTryInterrupt", ast.Load()), args, [])
        runTI = ast.Assign(
            [ast.Name(temporaryName, ast.Store())], ast.YieldFrom(callRuntime)
        )
        statements.append(runTI)
        result = ast.Name(temporaryName, ast.Load())
        if self.usedBreak:
            test = ast.Compare(result, [ast.Is()], [breakFlag])
            brk = ast.copy_location(ast.Break(), self.usedBreak)
            statements.append(ast.If(test, [brk], []))
        if self.usedContinue:
            test = ast.Compare(result, [ast.Is()], [continueFlag])
            cnt = ast.copy_location(ast.Continue(), self.usedContinue)
            statements.append(ast.If(test, [cnt], []))
        test = ast.Compare(result, [ast.Is()], [returnFlag])
        retCheck = ast.If(
            test, [ast.Return(ast.Attribute(result, "return_value", ast.Load()))], []
        )
        statements.append(retCheck)

        # Construct overall try-except statement
        if node.except_handlers or node.finalbody:
            newTry = ast.Try(
                statements,
                [self.visit(handler) for handler in node.except_handlers],
                self.visit(node.orelse),
                self.visit(node.finalbody),
            )
            return ast.copy_location(newTry, node)
        else:
            return statements

    def visit_TrackedAssign(self, node: s.TrackedAssign):
        return ast.Expr(
            value=ast.Call(
                func=ast.Name(id=node.target.functionName, ctx=loadCtx),
                args=[self.visit(node.value)],
                keywords=[],
            )
        )

    def visit_InitialScenario(self, node: s.InitialScenario):
        return ast.copy_location(
            ast.Call(
                func=ast.Name(id="in_initial_scenario", ctx=loadCtx),
                args=[],
                keywords=[],
            ),
            node,
        )

    def visit_ClassDef(self, node: ast.ClassDef) -> Any:
        # use `Object` as base if none is specified
        if not node.bases:
            node.bases = [ast.Name("Object", loadCtx)]

        # annotated assignments are not allowed in Scenic classes
        # those should be parsed as property definitions instead
        for stmt in node.body:
            if isinstance(stmt, ast.AnnAssign):
                raise self.makeSyntaxError(
                    "annotated assignments are not allowed in Scenic classes", stmt
                )

        # extract all property definitions
        propertyDefs: List[s.PropertyDef] = []
        newBody = []
        propSpot = None
        for stmt in node.body:
            if isinstance(stmt, s.PropertyDef):
                propertyDefs.append(stmt)
                if propSpot is None:
                    propSpot = len(newBody)
            else:
                newBody.append(stmt)

        # create dictionary from property name (str) to default values
        propertyDict = {}
        for propertyDef in propertyDefs:
            if propertyDef.property in propertyDict:
                raise self.makeSyntaxError(
                    f'duplicated property "{propertyDef.property}"', propertyDef
                )
            propertyDict[propertyDef.property] = propertyDef

        newBody.insert(
            len(newBody) if propSpot is None else propSpot,
            ast.Assign(
                targets=[ast.Name(id="_scenic_properties", ctx=ast.Store())],
                value=ast.Dict(
                    keys=[ast.Constant(value=p) for p in propertyDict.keys()],
                    values=[self.transformPropertyDef(v) for v in propertyDict.values()],
                ),
            ),
        )

        node.body = newBody
        return self.generic_visit(node)

    def transformPropertyDef(self, node: s.PropertyDef):
        properties, rawLoc = AttributeFinder.find("self", node.value)
        if rawLoc:
            self.makeSyntaxError(
                'cannot use raw name "self" in a default value (only "self.property")',
                rawLoc,
            )
        return ast.Call(
            func=ast.Name(id="_scenic_default", ctx=ast.Load()),
            args=[
                ast.Set(elts=[ast.Constant(value=p) for p in properties]),
                ast.Set(
                    elts=[ast.Constant(value=attr.keyword) for attr in node.attributes]
                ),
                ast.Lambda(
                    args=selfArg,
                    body=node.value,
                ),
            ],
            keywords=[],
        )

    def visit_PropertyDef(self, _: s.PropertyDef) -> Any:
        assert False, "PropertyDef should be handled in `visit_ClassDef`"

    @context(Context.TOP_LEVEL)
    def visit_BehaviorDef(self, node: s.BehaviorDef):
        return self.makeBehaviorLikeDef(
            baseClassName="Behavior",
            name=node.name,
            args=node.args,
            docstring=node.docstring,
            header=node.header,
            body=node.body,
        )

    @context(Context.TOP_LEVEL)
    def visit_MonitorDef(self, node: s.MonitorDef):
        return self.makeBehaviorLikeDef(
            baseClassName="Monitor",
            name=node.name,
            args=node.args,
            docstring=node.docstring,
            header=[],
            body=node.body,
        )

    @context(Context.TOP_LEVEL)
    def visit_ScenarioDef(self, node: s.ScenarioDef):
        # Set up arguments for setup and compose blocks
        args: ast.arguments = self.visit(node.args)
        args.posonlyargs = initialBehaviorArgs + args.posonlyargs

        # Get preconditions and invariants
        preconditions, invariants = self.separatePreconditionsAndInvariants(node.header)

        # Find all locals of the scenario, which will be shared amongst the various blocks
        allLocals = set()
        if node.compose:
            allLocals.update(LocalFinder.findIn(node.compose))
        if node.setup:
            allLocals.update(LocalFinder.findIn(node.setup))
        oldBL = self.behaviorLocals
        self.behaviorLocals = allLocals

        # Construct compose block
        self.inCompose = True
        guardCheckers = self.makeGuardCheckers(args, preconditions, invariants)
        if node.compose or preconditions or invariants:
            if node.compose:
                body = self.visit(node.compose)
            else:
                # generate no-op compose block to ensure invariants are checked
                wait = self.generateInvocation(node, ast.Constant(()))
                body = [ast.While(ast.Constant(True), wait, [])]
            compose = ast.FunctionDef("_compose", args, body, [], None)
        else:
            compose = ast.Assign([ast.Name("_compose", ast.Store())], ast.Constant(None))
        self.inCompose = False

        # Construct setup block
        self.inSetup = True
        if node.setup:
            setup = ast.FunctionDef("_setup", args, self.visit(node.setup), [], None)
        else:
            setup = ast.Assign([ast.Name("_setup", ast.Store())], ast.Constant(None))
        self.inSetup = False

        self.behaviorLocals = oldBL

        # Assemble scenario definition
        locs = ast.Constant(tuple(sorted(allLocals)))  # sort for AST determinism
        locs = ast.Call(ast.Name("frozenset", ast.Load()), [locs], [])
        saveLocals = ast.Assign([ast.Name("_locals", ast.Store())], locs)
        body = guardCheckers + [saveLocals, setup, compose]
        return ast.ClassDef(
            node.name, [ast.Name("DynamicScenario", loadCtx)], [], body, []
        )

    def makeGuardCheckers(
        self,
        args: ast.arguments,
        preconditions: List[s.Precondition],
        invariants: List[s.Invariant],
    ) -> List[ast.AST]:
        """Create a list of statements that defines precondition and invariant checker"""

        # Statements that check preconditions are satisfied
        preconditionChecks = []
        for precondition in preconditions:
            call = ast.Call(
                ast.Name("PreconditionViolation", loadCtx),
                [
                    ast.Name(behaviorArgName, loadCtx),
                    ast.Constant(precondition.lineno),
                ],
                [],
            )
            throw = ast.Raise(exc=call, cause=None)
            check = ast.If(
                test=ast.UnaryOp(ast.Not(), self.visit(precondition.value)),
                body=[throw],
                orelse=[],
            )

            chained_throw = ast.Raise(exc=call, cause=ast.Name("e", loadCtx))

            catch = ast.ExceptHandler(
                type=ast.Name("RejectionException", loadCtx),
                name="e",
                body=[chained_throw],
            )

            wrapped_check = ast.Try(
                body=[check], handlers=[catch], orelse=[], finalbody=[]
            )

            preconditionChecks.append(ast.copy_location(wrapped_check, precondition))

        definePreconditionChecker = ast.FunctionDef(
            checkPreconditionsName, args, preconditionChecks or [ast.Pass()], [], None
        )

        # Statements that check invariants are satisfied
        invariantChecks = []
        for invariant in invariants:
            call = ast.Call(
                ast.Name("InvariantViolation", loadCtx),
                [ast.Name(behaviorArgName, loadCtx), ast.Constant(invariant.lineno)],
                [],
            )
            throw = ast.Raise(exc=call, cause=None)
            check = ast.If(
                test=ast.UnaryOp(ast.Not(), self.visit(invariant.value)),
                body=[throw],
                orelse=[],
            )

            chained_throw = ast.Raise(exc=call, cause=ast.Name("e", loadCtx))

            catch = ast.ExceptHandler(
                type=ast.Name("RejectionException", loadCtx),
                name="e",
                body=[chained_throw],
            )

            wrapped_check = ast.Try(
                body=[check], handlers=[catch], orelse=[], finalbody=[]
            )

            invariantChecks.append(ast.copy_location(wrapped_check, invariant))

        defineInvariantChecker = ast.FunctionDef(
            checkInvariantsName, args, invariantChecks or [ast.Pass()], [], None
        )

        # assemble function body preamble
        preamble = [
            definePreconditionChecker,
            defineInvariantChecker,
        ]
        return preamble

    def separatePreconditionsAndInvariants(
        self, header: List[Union[s.Precondition, s.Invariant]]
    ) -> Tuple[List[s.Precondition], List[s.Invariant]]:
        """Given a list of preconditions and invariants, separate items into the list of preconditions and list of invariants

        Args:
            header (List[Union[s.Precondition, s.Invariant]]): List of preconditions and invariants

        Returns:
            Tuple[List[s.Precondition], List[s.Invariant]]: Tuple of precondition list and invariant list
        """
        preconditions: List[s.Precondition] = []
        invariants: List[s.Invariant] = []
        for n in header:
            if isinstance(n, s.Precondition):
                preconditions.append(n)
            elif isinstance(n, s.Invariant):
                invariants.append(n)
            else:
                assert False, f"Unexpected node type {n.__class__.__name__}"
        return (preconditions, invariants)

    def makeBehaviorLikeDef(
        self,
        baseClassName: Literal["Behavior", "Monitor"],
        name: str,
        args: Optional[ast.arguments],
        docstring: Optional[str],
        header: List[Union[s.Precondition, s.Invariant]],
        body: List[ast.AST],
    ):
        if baseClassName == "Behavior":
            ctxFlag = "inBehavior"
        elif baseClassName == "Monitor":
            ctxFlag = "inMonitor"
        else:
            assert False, f'Unexpected base class name "{baseClassName}"'

        # --- Extract preconditions and invariants ---
        preconditions, invariants = self.separatePreconditionsAndInvariants(header)

        # --- Copy arguments to the behavior object's namespace ---
        # list of all arguments
        allArgs = itertools.chain(args.posonlyargs, args.args, args.kwonlyargs)
        # statements that create argument variables
        copyArgs: List[ast.AST] = []
        for arg in allArgs:
            dest = ast.Attribute(ast.Name(behaviorArgName, loadCtx), arg.arg, ast.Store())
            copyArgs.append(
                ast.copy_location(ast.Assign([dest], ast.Name(arg.arg, loadCtx)), arg)
            )

        # --- Create a new `arguments` ---
        newArgs: ast.arguments = self.visit(args)
        # add private current behavior argument and implicit `self` argument
        newArgs.posonlyargs = initialBehaviorArgs + newArgs.posonlyargs

        # --- Process body ---
        setattr(self, ctxFlag, True)
        oldBehaviorLocals = self.behaviorLocals

        self.behaviorLocals = allLocals = LocalFinder.findIn(body)

        # handle docstring
        newBody = body
        docstringNode = []
        if docstring is not None:
            docstringNode = [ast.Expr(ast.Constant(unquote(docstring)))]

        # process body
        statements = self.visit(newBody)

        generatorBody = copyArgs + statements
        generatorDefinition = ast.FunctionDef(
            name="makeGenerator",
            args=newArgs,
            body=generatorBody,
            decorator_list=[],
            returns=None,
        )

        # --- Save local variables ---
        locs = ast.Constant(tuple(sorted(allLocals)))  # sort for AST determinism
        locs = ast.Call(ast.Name("frozenset", ast.Load()), [locs], [])
        saveLocals = ast.Assign([ast.Name("_locals", ast.Store())], locs)

        # --- Guards ---
        guardCheckers = self.makeGuardCheckers(newArgs, preconditions, invariants)

        # --- Create class definition ---
        classBody = docstringNode + guardCheckers + [saveLocals, generatorDefinition]
        classDefinition = ast.ClassDef(
            name, [ast.Name(baseClassName, loadCtx)], [], classBody, []
        )

        setattr(self, ctxFlag, False)
        self.behaviorLocals = oldBehaviorLocals

        return classDefinition

    def visit_Call(self, node: ast.Call) -> Any:
        newArgs = []
        wrappedStar = False
        for arg in node.args:
            if isinstance(arg, ast.Starred) and not self.inBehavior:
                wrappedStar = True
                checkedVal = ast.Call(
                    ast.Name("wrapStarredValue", ast.Load()),
                    [self.visit(arg.value), ast.Constant(arg.value.lineno)],
                    [],
                )
                newArgs.append(ast.Starred(checkedVal, ast.Load()))
            else:
                newArgs.append(self.visit(arg))
        newKeywords = [self.visit(kwarg) for kwarg in node.keywords]
        newFunc = self.visit(node.func)
        if wrappedStar:
            newNode = ast.Call(
                ast.Name("callWithStarArgs", ast.Load()),
                [newFunc] + newArgs,
                newKeywords,
            )
        else:
            newNode = ast.Call(newFunc, newArgs, newKeywords)
        newNode = ast.copy_location(newNode, node)
        return newNode

    @context(Context.TOP_LEVEL)
    def visit_Model(self, node: s.Model):
        if self.inSetup:
            raise self.makeSyntaxError('Cannot use "model" inside a setup block', node)
        return ast.Expr(
            value=ast.Call(
                func=ast.Name(id="model", ctx=loadCtx),
                args=[
                    ast.Name(id="_Scenic_module_namespace", ctx=loadCtx),
                    ast.Constant(value=node.name),
                ],
                keywords=[],
            )
        )

    @context(Context.TOP_LEVEL)
    def visit_Mutate(self, node: s.Mutate):
        return ast.Expr(
            value=ast.Call(
                func=ast.Name(id="mutate", ctx=loadCtx),
                args=[self.visit(el) for el in node.elts],
                keywords=(
                    [ast.keyword(arg="scale", value=self.visit(node.scale))]
                    if node.scale is not None
                    else []
                ),
            )
        )

    @context(Context.TOP_LEVEL)
    def visit_Param(self, node: s.Param):
        d = dict()
        for parameter in node.elts:
            if parameter.identifier in d:
                raise self.makeSyntaxError(
                    f'Duplicated param "{parameter.identifier}"', node
                )
            d[parameter.identifier] = self.visit(parameter.value)
        return ast.Expr(
            value=ast.Call(
                func=ast.Name(id="param", ctx=loadCtx),
                args=[
                    ast.Dict(
                        [ast.Constant(k) for k in d.keys()],
                        list(d.values()),
                    )
                ],
                keywords=[],
            )
        )

    def visit_Require(self, node: s.Require):
        prob = node.prob
        if prob is not None and not 0 <= prob <= 1:
            raise self.makeSyntaxError("probability must be between 0 and 1", node)
        return self.createRequirementLike(
            "require", node.cond, node.lineno, node.name, prob
        )

    def visit_RequireMonitor(self, node: s.RequireMonitor):
        return self.createRequirementLike(
            "require_monitor", node.monitor, node.lineno, node.name
        )

    def visit_Override(self, node: s.Override):
        return ast.Expr(
            value=ast.Call(
                func=ast.Name(id="override", ctx=loadCtx),
                args=[self.visit(node.target)] + [self.visit(s) for s in node.specifiers],
                keywords=[],
            )
        )

    def visit_Abort(self, node: s.Abort):
        if not self.inInterruptBlock:
            raise self.makeSyntaxError(
                "`abort` can only be used inside an interrupt block", node
            )
        return ast.copy_location(
            ast.Return(abortFlag),
            node,
        )

    @context(Context.BEHAVIOR)
    def visit_Take(self, node: s.Take):
        action = ast.Tuple(self.visit(node.elts), loadCtx)
        return self.generateInvocation(node, ast.copy_location(action, node))

    @context(Context.DYNAMIC)
    def visit_Wait(self, node: s.Wait):
        return self.generateInvocation(node, ast.Constant(()))

    @context(Context.DYNAMIC)
    def visit_Terminate(self, node: s.Terminate):
        termination = ast.Call(
            ast.Name("_makeTerminationAction", loadCtx),
            [ast.Name("self", ast.Load()), ast.Constant(node.lineno)],
            [],
        )
        return self.generateInvocation(node, termination)

    @context(Context.DYNAMIC)
    def visit_TerminateSimulation(self, node: s.TerminateSimulation):
        termination = ast.Call(
            ast.Name("_makeSimulationTerminationAction", loadCtx),
            [ast.Constant(node.lineno)],
            [],
        )
        return self.generateInvocation(node, termination)

    @context(Context.DYNAMIC)
    def visit_Do(self, node: s.Do):
        if (self.inBehavior or self.inMonitor) and len(node.elts) > 1:
            raise self.makeSyntaxError(
                f"`do` can only take one action inside a {'behavior' if self.inBehavior else 'monitor'}",
                node,
            )
        return self.makeDoLike(node, node.elts)

    @context(Context.DYNAMIC)
    def visit_DoFor(self, node: s.DoFor):
        if (self.inBehavior or self.inMonitor) and len(node.elts) > 1:
            raise self.makeSyntaxError(
                f"`do` can only take one action inside a {'behavior' if self.inBehavior else 'monitor'}",
                node,
            )
        return self.makeDoLike(
            node,
            node.elts,
            modifier=ast.Call(
                func=ast.Name("Modifier", loadCtx),
                args=[
                    ast.Constant("for"),
                    self.visit(node.duration.value),
                    ast.Constant(node.duration.unitStr),
                ],
                keywords=[],
            ),
        )

    @context(Context.DYNAMIC)
    def visit_DoUntil(self, node: s.DoUntil):
        if (self.inBehavior or self.inMonitor) and len(node.elts) > 1:
            raise self.makeSyntaxError(
                f"`do` can only take one action inside a {'behavior' if self.inBehavior else 'monitor'}",
                node,
            )
        return self.makeDoLike(
            node,
            node.elts,
            modifier=ast.Call(
                func=ast.Name("Modifier", loadCtx),
                args=[
                    ast.Constant("until"),
                    ast.Lambda(noArgs, self.visit(node.cond)),
                ],
                keywords=[],
            ),
        )

    def visit_DoChoose(self, node: s.DoChoose):
        return self.makeDoLike(node, node.elts, schedule="choose")

    def visit_DoShuffle(self, node: s.DoChoose):
        return self.makeDoLike(node, node.elts, schedule="shuffle")

    def generateInvocation(self, node: ast.AST, actionlike, invoker=ast.Yield):
        """Generate an invocation of an action, behavior, or scenario."""
        invokeAction = ast.Expr(invoker(actionlike))
        checker = ast.Attribute(
            ast.Name(behaviorArgName, loadCtx), checkInvariantsName, loadCtx
        )
        args = ast.Starred(
            ast.Attribute(ast.Name(behaviorArgName, loadCtx), "_args", loadCtx), loadCtx
        )
        kwargs = ast.keyword(
            None, ast.Attribute(ast.Name(behaviorArgName, loadCtx), "_kwargs", loadCtx)
        )
        checkInvariants = ast.Expr(
            ast.Call(checker, [ast.Name("self", loadCtx), args], [kwargs])
        )
        ast.copy_location(invokeAction, node)
        ast.copy_location(checkInvariants, node)
        return [
            invokeAction,
            checkInvariants,
        ]

    def makeDoLike(
        self,
        node: ast.AST,
        elts: List[ast.AST],
        modifier: Optional[ast.Call] = None,
        schedule: Optional[str] = None,
    ):
        subHandler = ast.Attribute(
            ast.Name(behaviorArgName, loadCtx), "_invokeSubBehavior", loadCtx
        )
        subArgs = [
            ast.Name("self", loadCtx),
            ast.Tuple([self.visit(e) for e in elts], loadCtx),
        ]
        if modifier is not None:
            subArgs.append(modifier)
        keywords = []
        if schedule is not None:
            keywords = [ast.keyword("schedule", ast.Constant(schedule))]
        subRunner = ast.Call(subHandler, subArgs, keywords)
        return self.generateInvocation(node, subRunner, ast.YieldFrom)

    @context(Context.TOP_LEVEL)
    def visit_Record(self, node: s.Record):
        return self.createRequirementLike("record", node.value, node.lineno, node.name)

    @context(Context.TOP_LEVEL)
    def visit_RecordInitial(self, node: s.RecordInitial):
        return self.createRequirementLike(
            "record_initial", node.value, node.lineno, node.name
        )

    @context(Context.TOP_LEVEL)
    def visit_RecordFinal(self, node: s.RecordFinal):
        return self.createRequirementLike(
            "record_final", node.value, node.lineno, node.name
        )

    @context(Context.TOP_LEVEL)
    def visit_TerminateWhen(self, node: s.TerminateWhen):
        return self.createRequirementLike(
            "terminate_when", node.cond, node.lineno, node.name
        )

    @context(Context.TOP_LEVEL)
    def visit_TerminateSimulationWhen(self, node: s.TerminateSimulationWhen):
        return self.createRequirementLike(
            "terminate_simulation_when", node.cond, node.lineno, node.name
        )

    def createRequirementLike(
        self,
        functionName: str,
        body: ast.AST,
        lineno: int,
        name: Optional[str] = None,
        prob: Optional[float] = None,
    ):
        """Create a call to a function that implements requirement-like features, such as `record` and `terminate when`.

        Args:
            functionName (str): Name of the requirement-like function to call. Its signature must be `(reqId: int, body: () -> bool, lineno: int, name: str | None)`
            body (ast.AST): AST node to evaluate for checking the condition
            lineno (int): Line number in the source code
            name (Optional[str], optional): Optional name for requirements. Defaults to None.
            prob (Optional[float], optional): Optional probability for requirements. Defaults to None.
        """
        propTransformer = PropositionTransformer(self.filename)
        newBody, self.nextSyntaxId = propTransformer.transform(body, self.nextSyntaxId)
        newBody = self.visit(newBody)
        requirementId = self._register_requirement_syntax(body)

        return ast.Expr(
            value=ast.Call(
                func=ast.Name(functionName, loadCtx),
                args=[
                    ast.Constant(requirementId),  # requirement IDre
                    newBody,  # body
                    ast.Constant(lineno),  # line number
                    ast.Constant(name),  # requirement name
                ],
                keywords=(
                    [ast.keyword(arg="prob", value=ast.Constant(prob))]
                    if prob is not None
                    else []
                ),
            )
        )

    @context(Context.TOP_LEVEL)
    def visit_TerminateAfter(self, node: s.TerminateAfter):
        return ast.copy_location(
            ast.Expr(
                ast.Call(
                    func=ast.Name(id="terminate_after", ctx=loadCtx),
                    args=[
                        self.visit(node.duration.value),
                        ast.Constant(node.duration.unitStr),
                    ],
                    keywords=[],
                )
            ),
            node,
        )

    @context(Context.TOP_LEVEL)
    def visit_Simulator(self, node: s.Simulator):
        return ast.copy_location(
            ast.Expr(
                ast.Call(
                    func=ast.Name(id="simulator", ctx=loadCtx),
                    args=[ast.Lambda(noArgs, self.visit(node.value))],
                    keywords=[],
                )
            ),
            node,
        )

    # Instance & Specifier

    def visit_New(self, node: s.New):
        return ast.Call(
            func=ast.Name(id="new", ctx=loadCtx),
            args=[
                ast.Name(id=node.className, ctx=loadCtx),
                ast.List(elts=[self.visit(s) for s in node.specifiers], ctx=ast.Load()),
            ],
            keywords=[],
        )

    def visit_WithSpecifier(self, node: s.WithSpecifier):
        return ast.Call(
            func=ast.Name(id="With", ctx=loadCtx),
            args=[
                ast.Constant(value=node.prop),
                self.visit(node.value),
            ],
            keywords=[],
        )

    def visit_AtSpecifier(self, node: s.AtSpecifier):
        return ast.Call(
            func=ast.Name(id="At", ctx=loadCtx),
            args=[self.visit(node.position)],
            keywords=[],
        )

    def visit_OffsetBySpecifier(self, node: s.OffsetBySpecifier):
        return ast.Call(
            func=ast.Name(id="OffsetBy", ctx=loadCtx),
            args=[
                self.visit(node.offset),
            ],
            keywords=[],
        )

    def visit_OffsetAlongSpecifier(self, node: s.OffsetAlongSpecifier):
        return ast.Call(
            func=ast.Name(id="OffsetAlongSpec", ctx=loadCtx),
            args=[
                self.visit(node.direction),
                self.visit(node.offset),
            ],
            keywords=[],
        )

    def visit_DirectionOfSpecifier(self, node: s.DirectionOfSpecifier):
        if isinstance(node.direction, s.LeftOf):
            fn = "LeftSpec"
        elif isinstance(node.direction, s.RightOf):
            fn = "RightSpec"
        elif isinstance(node.direction, s.AheadOf):
            fn = "Ahead"
        elif isinstance(node.direction, s.Behind):
            fn = "Behind"
        elif isinstance(node.direction, s.Above):
            fn = "Above"
        elif isinstance(node.direction, s.Below):
            fn = "Below"
        else:
            assert False, f"impossible direction {node.direction} in PositionSpecifier"
        return ast.Call(
            func=ast.Name(id=fn, ctx=loadCtx),
            args=[
                self.visit(node.position),
            ],
            keywords=(
                []
                if node.distance is None
                else [ast.keyword(arg="dist", value=self.visit(node.distance))]
            ),
        )

    def visit_BeyondSpecifier(self, node: s.BeyondSpecifier):
        return ast.Call(
            func=ast.Name(id="Beyond", ctx=loadCtx),
            args=[self.visit(node.position), self.visit(node.offset)],
            keywords=(
                [ast.keyword(arg="fromPt", value=self.visit(node.base))]
                if node.base is not None
                else []
            ),
        )

    def visit_VisibleSpecifier(self, node: s.VisibleSpecifier):
        if node.base is not None:
            return ast.Call(
                func=ast.Name(id="VisibleFrom", ctx=loadCtx),
                args=[self.visit(node.base)],
                keywords=[],
            )
        return ast.Call(
            func=ast.Name(id="VisibleSpec", ctx=loadCtx),
            args=[],
            keywords=[],
        )

    def visit_NotVisibleSpecifier(self, node: s.NotVisibleSpecifier):
        if node.base is not None:
            return ast.Call(
                func=ast.Name(id="NotVisibleFrom", ctx=loadCtx),
                args=[self.visit(node.base)],
                keywords=[],
            )
        return ast.Call(
            func=ast.Name(id="NotVisibleSpec", ctx=loadCtx),
            args=[],
            keywords=[],
        )

    def visit_InSpecifier(self, node: s.InSpecifier):
        return ast.Call(
            func=ast.Name(id="In", ctx=loadCtx),
            args=[self.visit(node.region)],
            keywords=[],
        )

    def visit_OnSpecifier(self, node: s.InSpecifier):
        return ast.Call(
            func=ast.Name(id="On", ctx=loadCtx),
            args=[self.visit(node.region)],
            keywords=[],
        )

    def visit_ContainedInSpecifier(self, node: s.InSpecifier):
        return ast.Call(
            func=ast.Name(id="ContainedIn", ctx=loadCtx),
            args=[self.visit(node.region)],
            keywords=[],
        )

    def visit_FollowingSpecifier(self, node: s.FollowingSpecifier):
        return ast.Call(
            func=ast.Name(id="Following", ctx=loadCtx),
            args=[self.visit(node.field), self.visit(node.distance)],
            keywords=(
                [ast.keyword(arg="fromPt", value=self.visit(node.base))]
                if node.base is not None
                else []
            ),
        )

    def visit_FacingSpecifier(self, node: s.FacingSpecifier):
        return ast.Call(
            func=ast.Name(id="Facing", ctx=loadCtx),
            args=[self.visit(node.heading)],
            keywords=[],
        )

    def visit_FacingTowardSpecifier(self, node: s.FacingTowardSpecifier):
        return ast.Call(
            func=ast.Name(id="FacingToward", ctx=loadCtx),
            args=[self.visit(node.position)],
            keywords=[],
        )

    def visit_FacingAwayFromSpecifier(self, node: s.FacingAwayFromSpecifier):
        return ast.Call(
            func=ast.Name(id="FacingAwayFrom", ctx=loadCtx),
            args=[self.visit(node.position)],
            keywords=[],
        )

    def visit_FacingDirectlyTowardSpecifier(self, node: s.FacingDirectlyTowardSpecifier):
        return ast.Call(
            func=ast.Name(id="FacingDirectlyToward", ctx=loadCtx),
            args=[self.visit(node.position)],
            keywords=[],
        )

    def visit_FacingDirectlyAwayFromSpecifier(
        self, node: s.FacingDirectlyAwayFromSpecifier
    ):
        return ast.Call(
            func=ast.Name(id="FacingDirectlyAwayFrom", ctx=loadCtx),
            args=[self.visit(node.position)],
            keywords=[],
        )

    def visit_ApparentlyFacingSpecifier(self, node: s.ApparentlyFacingSpecifier):
        return ast.Call(
            func=ast.Name(id="ApparentlyFacing", ctx=loadCtx),
            args=[self.visit(node.heading)],
            keywords=(
                [ast.keyword(arg="fromPt", value=self.visit(node.base))]
                if node.base is not None
                else []
            ),
        )

    # Operators

    def visit_RelativePositionOp(self, node: s.RelativePositionOp):
        return ast.Call(
            func=ast.Name(id="RelativePosition", ctx=loadCtx),
            args=[self.visit(node.target)],
            keywords=(
                []
                if node.base is None
                else [ast.keyword(arg="Y", value=self.visit(node.base))]
            ),
        )

    def visit_RelativeHeadingOp(self, node: s.RelativeHeadingOp):
        return ast.Call(
            func=ast.Name(id="RelativeHeading", ctx=loadCtx),
            args=[self.visit(node.target)],
            keywords=(
                []
                if node.base is None
                else [ast.keyword(arg="Y", value=self.visit(node.base))]
            ),
        )

    def visit_ApparentHeadingOp(self, node: s.ApparentHeadingOp):
        return ast.Call(
            func=ast.Name(id="ApparentHeading", ctx=loadCtx),
            args=[self.visit(node.target)],
            keywords=(
                []
                if node.base is None
                else [ast.keyword(arg="Y", value=self.visit(node.base))]
            ),
        )

    def visit_DistanceFromOp(self, node: s.DistanceFromOp):
        return ast.Call(
            func=ast.Name(id="DistanceFrom", ctx=loadCtx),
            args=[self.visit(node.target)],
            keywords=(
                [ast.keyword(arg="Y", value=self.visit(node.base))]
                if node.base is not None
                else []
            ),
        )

    def visit_DistancePastOp(self, node: s.DistancePastOp):
        return ast.Call(
            func=ast.Name(id="DistancePast", ctx=loadCtx),
            args=[self.visit(node.target)],
            keywords=(
                []
                if node.base is None
                else [ast.keyword(arg="Y", value=self.visit(node.base))]
            ),
        )

    def visit_AngleFromOp(self, node: s.AngleFromOp):
        assert (
            node.base is not None or node.target is not None
        ), "neither target nor base were specified in AngleFromOp"
        keywords = []
        if node.base is not None:
            keywords.append(ast.keyword("X", self.visit(node.base)))
        if node.target is not None:
            keywords.append(ast.keyword("Y", self.visit(node.target)))
        return ast.Call(
            func=ast.Name(id="AngleFrom", ctx=loadCtx),
            args=[],
            keywords=keywords,
        )

    def visit_AltitudeFromOp(self, node: s.AltitudeFromOp):
        assert (
            node.base is not None or node.target is not None
        ), "neither target nor base were specified in AltitudeFromOp"
        keywords = []
        if node.base is not None:
            keywords.append(ast.keyword("X", self.visit(node.base)))
        if node.target is not None:
            keywords.append(ast.keyword("Y", self.visit(node.target)))
        return ast.Call(
            func=ast.Name(id="AltitudeFrom", ctx=loadCtx),
            args=[],
            keywords=keywords,
        )

    def visit_FollowOp(self, node: s.FollowOp):
        return ast.Call(
            func=ast.Name(id="Follow", ctx=loadCtx),
            args=[
                self.visit(node.target),
                self.visit(node.base),
                self.visit(node.distance),
            ],
            keywords=[],
        )

    def visit_VisibleOp(self, node: s.VisibleOp):
        return ast.Call(
            func=ast.Name(id="Visible", ctx=loadCtx),
            args=[self.visit(node.region)],
            keywords=[],
        )

    def visit_NotVisibleOp(self, node: s.VisibleOp):
        return ast.Call(
            func=ast.Name(id="NotVisible", ctx=loadCtx),
            args=[self.visit(node.region)],
            keywords=[],
        )

    def visit_VisibleFromOp(self, node: s.VisibleFromOp):
        return ast.Call(
            func=ast.Name(id="VisibleFromOp", ctx=loadCtx),
            args=[self.visit(node.region), self.visit(node.base)],
            keywords=[],
        )

    def visit_NotVisibleFromOp(self, node: s.VisibleFromOp):
        return ast.Call(
            func=ast.Name(id="NotVisibleFromOp", ctx=loadCtx),
            args=[self.visit(node.region), self.visit(node.base)],
            keywords=[],
        )

    def visit_PositionOfOp(self, node: s.PositionOfOp):
        return ast.Call(
            func=ast.Name(id=node.position.functionName, ctx=loadCtx),
            args=[self.visit(node.target)],
            keywords=[],
        )

    def visit_DegOp(self, node: s.DegOp):
        return ast.BinOp(
            left=self.visit(node.operand),
            op=ast.Mult(),
            right=ast.Constant(0.017453292519943295),
        )

    def visit_VectorOp(self, node: s.VectorOp):
        return ast.Call(
            func=ast.Name(id="Vector", ctx=loadCtx),
            args=[self.visit(node.left), self.visit(node.right)],
            keywords=[],
        )

    def visit_FieldAtOp(self, node: s.FieldAtOp):
        return ast.Call(
            func=ast.Name(id="FieldAt", ctx=loadCtx),
            args=[self.visit(node.left), self.visit(node.right)],
            keywords=[],
        )

    def visit_RelativeToOp(self, node: s.RelativeToOp):
        return ast.Call(
            func=ast.Name(id="RelativeTo", ctx=loadCtx),
            args=[self.visit(node.left), self.visit(node.right)],
            keywords=[],
        )

    def visit_OffsetAlongOp(self, node: s.OffsetAlongOp):
        return ast.Call(
            func=ast.Name(id="OffsetAlong", ctx=loadCtx),
            args=[
                self.visit(node.base),
                self.visit(node.direction),
                self.visit(node.offset),
            ],
            keywords=[],
        )

    def visit_CanSeeOp(self, node: s.CanSeeOp):
        return ast.Call(
            func=ast.Name(id="CanSee", ctx=loadCtx),
            args=[
                self.visit(node.left),
                self.visit(node.right),
            ],
            keywords=[],
        )

    def visit_IntersectsOp(self, node: s.IntersectsOp):
        return ast.Call(
            func=ast.Name(id="Intersects", ctx=loadCtx),
            args=[
                self.visit(node.left),
                self.visit(node.right),
            ],
            keywords=[],
        )



"""Preconditions and invariants of behaviors and scenarios."""


class GuardViolation(Exception):
    """Abstract exception raised when a guard of a behavior is violated.

    This will never be raised directly; either of the subclasses `PreconditionViolation`
    or `InvariantViolation` will be used, as appropriate.
    """

    violationType = "guard"

    def __init__(self, behavior, lineno):
        self.behaviorName = behavior.__class__.__name__
        self.lineno = lineno

    def __str__(self):
        return (
            f"violated {self.violationType} of {self.behaviorName} on line {self.lineno}"
        )


class PreconditionViolation(GuardViolation):
    """Exception raised when a precondition is violated.

    Raised when a precondition is violated when invoking a behavior
    or when a precondition encounters a `RejectionException`, so that
    rejections count as precondition violations.
    """

    violationType = "precondition"


class InvariantViolation(GuardViolation):
    """Exception raised when an invariant is violated.

    Raised when an invariant is violated when invoking/resuming a behavior
    or when an invariant encounters a `RejectionException`, so that
    rejections count as invariant violations.
    """

    violationType = "invariant"



"""Interface to Webots for dynamic simulations.

This interface is intended to be instantiated from inside the controller script
of a Webots `Robot node`_ with the ``supervisor`` field set to true. Such a
script can create a `WebotsSimulator` (passing in a reference to the supervisor
node) and then call its `simulate` method as usual to run a simulation. For an
example, see :file:`examples/webots/generic/controllers/scenic_supervisor.py`.

Scenarios written for this interface should use our generic Webots world model
:doc:`scenic.simulators.webots.model` or a model derived from it. Objects which
are instances of `WebotsObject` will be matched to Webots nodes; see the model
documentation for details.

.. _Robot node: https://www.cyberbotics.com/doc/reference/robot
"""

from collections import defaultdict
import ctypes
import math
from os import path
import tempfile
from textwrap import dedent

import trimesh

from scenic.core.regions import MeshVolumeRegion
from scenic.core.simulators import Simulation, Simulator
from scenic.core.type_support import toOrientation
from scenic.core.vectors import Vector
from scenic.simulators.webots.utils import ENU, WebotsCoordinateSystem


class WebotsSimulator(Simulator):
    """`Simulator` object for Webots.

    Args:
        supervisor: Supervisor node handle from the Webots Python API.
    """

    def __init__(self, supervisor):
        super().__init__()
        self.supervisor = supervisor
        topLevelNodes = supervisor.getRoot().getField("children")
        worldInfo = None
        for i in range(topLevelNodes.getCount()):
            child = topLevelNodes.getMFNode(i)
            if child.getTypeName() == "WorldInfo":
                worldInfo = child
                break
        if not worldInfo:
            raise RuntimeError("Webots world does not contain a WorldInfo node")
        system = worldInfo.getField("coordinateSystem").getSFString()
        self.coordinateSystem = WebotsCoordinateSystem(system)

    def createSimulation(self, scene, **kwargs):
        return WebotsSimulation(
            scene, self.supervisor, coordinateSystem=self.coordinateSystem, **kwargs
        )


class WebotsSimulation(Simulation):
    """`Simulation` object for Webots.

    Attributes:
        supervisor: Webots supervisor node used for the simulation. This is
            exposed for the use of scenarios which need to call Webots APIs
            directly; e.g. :scenic:`simulation().supervisor.setLabel({...})`.
    """

    def __init__(self, scene, supervisor, coordinateSystem=ENU, *, timestep, **kwargs):
        self.supervisor = supervisor
        self.coordinateSystem = coordinateSystem
        self.mode2D = scene.compileOptions.mode2D
        self.nextAdHocObjectId = 1
        self.usedObjectNames = defaultdict(lambda: 0)

        # directory to store proto files for adhoc webots objects
        self.tmpMeshDir = tempfile.mkdtemp()

        timestep = supervisor.getBasicTimeStep() / 1000 if timestep is None else timestep

        super().__init__(scene, timestep=timestep, **kwargs)

    def setup(self):
        super().setup()

        # Reset Webots simulation
        self.supervisor.simulationResetPhysics()

    def createObjectInSimulator(self, obj):
        if not hasattr(obj, "webotsName"):
            return  # not a Webots object

        # Find the name of the Webots node for this object.
        name = None
        if obj.webotsAdhoc is not None:
            # Dynamically generate object from Scenic object
            objectRawMesh = obj.shape.mesh
            objectScaledMesh = MeshVolumeRegion(
                mesh=objectRawMesh,
                dimensions=(obj.width, obj.length, obj.height),
            ).mesh
            objFilePath = path.join(self.tmpMeshDir, f"{self.nextAdHocObjectId}.obj")
            trimesh.exchange.export.export_mesh(objectScaledMesh, objFilePath)

            name = self._getAdhocObjectName(self.nextAdHocObjectId)
            protoName = (
                "ScenicObjectWithPhysics" if isPhysicsEnabled(obj) else "ScenicObject"
            )
            protoDef = dedent(
                f"""
                DEF {name} {protoName} {{
                    url "{objFilePath}"
                }}
                """
            )

            rootNode = self.supervisor.getRoot()
            rootChildrenField = rootNode.getField("children")
            rootChildrenField.importMFNodeFromString(-1, protoDef)
            self.nextAdHocObjectId += 1
        else:
            if obj.webotsName:
                name = obj.webotsName
            else:
                ty = obj.webotsType
                if not ty:
                    raise RuntimeError(f"object {obj} has no webotsName or webotsType")
                nextID = self.usedObjectNames[ty]
                self.usedObjectNames[ty] += 1
                if nextID == 0 and self.supervisor.getFromDef(ty):
                    name = ty
                else:
                    name = f"{ty}_{nextID}"

        # Get handle to Webots node.
        webotsObj = self.supervisor.getFromDef(name)
        if webotsObj is None:
            raise SimulationCreationError(f"Webots object {name} does not exist in world")
        obj.webotsObject = webotsObj
        obj.webotsName = name

        # Set the fields of the Webots object:

        # position
        if self.mode2D:  # 2D compatibility mode
            # Set initial elevation if unspecified
            if obj.elevation is None:
                pos = webotsObj.getField("translation").getSFVec3f()
                spos = self.coordinateSystem.positionToScenic(pos)
                obj.elevation = spos[2]

            # Overwrite Z value with elevation
            pos = self.coordinateSystem.positionFromScenic(
                Vector(obj.position.x, obj.position.y, obj.elevation) + obj.positionOffset
            )
            webotsObj.getField("translation").setSFVec3f(pos)
        else:
            pos = self.coordinateSystem.positionFromScenic(
                obj.position + obj.positionOffset
            )
            webotsObj.getField("translation").setSFVec3f(pos)

        # orientation
        offsetOrientation = toOrientation(obj.rotationOffset)
        webotsObj.getField("rotation").setSFRotation(
            self.coordinateSystem.orientationFromScenic(
                obj.orientation, offsetOrientation
            )
        )

        # density
        densityField = getFieldSafe(webotsObj, "density")
        if densityField is not None:
            if obj.density is None:
                # Get initial value for property if unspecified
                obj.density = densityField.getSFFloat()
            else:
                densityField.setSFFloat(float(obj.density))

        # battery
        battery = getattr(obj, "battery", None)
        if battery:
            if not isinstance(battery, (tuple, list)) or len(battery) != 3:
                raise TypeError(f'"battery" of {name} does not have 3 components')
            field = webotsObj.getField("battery")
            field.setMFFloat(0, battery[0])
            field.setMFFloat(1, battery[1])
            field.setMFFloat(2, battery[2])

        # customData
        customData = getattr(obj, "customData", None)
        if customData:
            if not isinstance(customData, str):
                raise TypeError(f'"customData" of {name} is not a string')
            webotsObj.getField("customData").setSFString(customData)

        # controller
        if obj.controller:
            controllerField = webotsObj.getField("controller")
            curCont = controllerField.getSFString()
            if obj.controller != curCont:
                # the following operation also causes the controller to be restarted
                controllerField.setSFString(obj.controller)
            elif obj.resetController:
                webotsObj.restartController()

    def step(self):
        ms = round(1000 * self.timestep)
        self.supervisor.step(ms)

    def getProperties(self, obj, properties):
        webotsObj = getattr(obj, "webotsObject", None)
        if not webotsObj:  # static object with no Webots counterpart
            return {prop: getattr(obj, prop) for prop in properties}

        pos = webotsObj.getField("translation").getSFVec3f()
        x, y, z = self.coordinateSystem.positionToScenic(pos)
        lx, ly, lz, ax, ay, az = webotsObj.getVelocity()
        vx, vy, vz = self.coordinateSystem.positionToScenic((lx, ly, lz))
        velocity = Vector(vx, vy, vz)
        speed = math.hypot(*velocity)
        angularSpeed = math.hypot(ax, ay, az)

        offsetOrientation = toOrientation(obj.rotationOffset)
        globalOrientation = self.coordinateSystem.orientationToScenic(
            webotsObj.getField("rotation").getSFRotation(), offsetOrientation
        )
        yaw, pitch, roll = obj.parentOrientation.localAnglesFor(globalOrientation)

        values = dict(
            position=Vector(x, y, z),
            velocity=velocity,
            speed=speed,
            angularSpeed=angularSpeed,
            angularVelocity=Vector(ax, ay, az),
            yaw=yaw,
            pitch=pitch,
            roll=roll,
            elevation=z,
        )

        if hasattr(obj, "battery"):
            field = webotsObj.getField("battery")
            val = (field.getMFFloat(0), obj.battery[1], obj.battery[2])
            values["battery"] = val

        return values

    def destroy(self):
        # Destroy adhoc objects generated at the beginning of the simulation
        for i in range(1, self.nextAdHocObjectId):
            name = self._getAdhocObjectName(i)
            node = self.supervisor.getFromDef(name)
            node.remove()

    def _getAdhocObjectName(self, i: int) -> str:
        return f"SCENIC_ADHOC_{i}"


def getFieldSafe(webotsObject, fieldName):
    """Get field from webots object. Return null if no such field exists.

    Needed to workaround this issue (https://github.com/cyberbotics/webots/issues/5646)

    Args:
        webotsObject: webots object
        fieldName: name of the field to look for

    Returns:
        Field|None: Field object if the field with the given name exists. None otherwise.
    """

    field = webotsObject.getField(fieldName)
    # this seems to always return some object, but return None if field is None
    if field is None:
        return None

    # if field is valid, it has a valid pointer
    if isinstance(field._ref, ctypes.c_void_p) and field._ref.value is not None:
        # then the field is valid and we return the reference
        return field

    # if the pointer points to None, then the field does not exist on this object
    return None


def isPhysicsEnabled(webotsObject):
    """Whether or not physics is enabled for this `WebotsObject`"""
    if webotsObject.webotsAdhoc is None:
        return webotsObject
    if isinstance(webotsObject.webotsAdhoc, dict):
        return webotsObject.webotsAdhoc.get("physics", True)
    raise TypeError(f"webotsAdhoc must be None or a dictionary")



"""The SampleChecker class and it's implementations."""

from abc import ABC, abstractmethod
from collections import deque
import time

from scenic.core.distributions import RejectionException
from scenic.core.requirements import BlanketCollisionRequirement, IntersectionRequirement


class SampleChecker(ABC):
    def __init__(self):
        self.requirements = None

    def setRequirements(self, requirements):
        assert self.requirements is None
        self.requirements = tuple(requirements)

    @abstractmethod
    def checkRequirementsInner(self, sample):
        pass

    def checkRequirements(self, sample):
        assert self.requirements is not None
        try:
            return self.checkRequirementsInner(sample)
        except RejectionException as e:
            return e


class BasicChecker(SampleChecker):
    """Basic requirement checker.

    Evaluates requirements in order, with a tiny bit of tuning.
    """

    def __init__(self, initialCollisionCheck):
        super().__init__()
        self.initialCollisionCheck = initialCollisionCheck

    def setRequirements(self, requirements):
        target_reqs = []
        for req in requirements:
            if req.optional:
                # Basic checker ignores optional requirements unless otherwise noted.
                if (
                    isinstance(req, BlanketCollisionRequirement)
                    and self.initialCollisionCheck
                    and sum(isinstance(r, IntersectionRequirement) for r in requirements)
                    >= 3
                ):
                    target_reqs.append(req)
            else:
                target_reqs.append(req)

        super().setRequirements(target_reqs)

    def checkRequirementsInner(self, sample):
        for req in self.requirements:
            if req.active and req.falsifiedBy(sample):
                return req.violationMsg

        return None


class WeightedAcceptanceChecker(SampleChecker):
    """Picks the requirement with the lowest time-weighted acceptance chance.

    Incentivizes exploration by initializing all buffer values to 0.

    Args:
        bufferSize: Max samples to use when calculating time-weighted
            rejection chance.
    """

    def __init__(self, bufferSize=10):
        super().__init__()
        self.bufferSize = bufferSize
        self.buffers = None
        self.bufferSums = None

    def setRequirements(self, requirements):
        super().setRequirements(requirements)

        self.buffers = {req: deque() for req in self.requirements}
        self.bufferSums = {req: (0, 0) for req in self.requirements}
        for req in self.requirements:
            self.buffers[req].extend([(0, 0)] * self.bufferSize)

    def checkRequirementsInner(self, sample):
        for req in self.sortedRequirements():
            # Evaluate the requirement with timing info.
            start = time.perf_counter()
            rejected = req.falsifiedBy(sample)
            # Create metrics (Accepted, Time Taken)
            metrics = (int(not rejected), time.perf_counter() - start)

            self.updateMetrics(req, metrics)

            if rejected:
                return req.violationMsg

        return None

    def sortedRequirements(self):
        """Return the list of requirements in sorted order"""
        # Extract and sort active requirements
        reqs = [req for req in self.requirements if req.active]
        reqs.sort(key=self.getWeightedAcceptanceProb)

        # Remove any optional requirements at the end of the list, since they're useless
        while reqs and reqs[-1].optional:
            reqs.pop()

        return reqs

    def updateMetrics(self, req, new_metrics):
        """Update the metrics for a given requirement"""
        # Update buffer
        target_buffer = self.buffers[req]
        old_metrics = target_buffer.popleft()
        target_buffer.append(new_metrics)

        # Unpack values
        sum_acc, sum_time = self.bufferSums[req]
        old_acc, old_time = old_metrics
        new_acc, new_time = new_metrics

        # Update sums
        sum_acc += new_acc - old_acc
        sum_time += new_time - old_time
        self.bufferSums[req] = (sum_acc, sum_time)

    def getWeightedAcceptanceProb(self, req):
        sum_acc, sum_time = self.bufferSums[req]
        return (sum_acc / self.bufferSize) * (sum_time / self.bufferSize)



"""
This file has basic image modification functions
"""

from itertools import product

from PIL import Image
import cv2
import numpy as np


def convert_black_white(img_data=None, img_file=None, threshold=100):
    assert img_data is not None or img_file is not None
    if img_data is None:
        img_data = Image.open(img_file)

    img_copy = img_data.copy()
    pixels = img_copy.load()

    for j, k in product(range(img_copy.size[0]), range(img_copy.size[1])):
        if (np.array(pixels[j, k][0:3]) > threshold).any():
            pixels[j, k] = (255, 255, 255, 255)
        else:
            pixels[j, k] = (0, 0, 0, 255)

    return img_copy


def get_edges(img_data=None, img_file=None, threshold=100, kernelsize=1):
    assert img_data is not None or img_file is not None
    if img_data is None:
        img_data = Image.open(img_file)

    img_copy = img_data.copy()

    # Get the black and white image
    img_bw = convert_black_white(
        img_data=img_copy, img_file=img_file, threshold=threshold
    )
    cv_bw = cv2.cvtColor(np.array(img_bw), cv2.COLOR_RGB2BGR)
    # Detect edges using Laplacian
    laplacian = cv2.Laplacian(cv_bw, cv2.CV_8U, ksize=kernelsize)

    # Convert back to Pillow image
    pil_lap = Image.fromarray(laplacian)

    # For computing Voronoi images, we need to squeeze the RGB data to 0s and 1s
    pil_squeezed = pil_lap.convert("L")
    pil_squeezed_01 = pil_squeezed.point(lambda x: 0 if x < 128 else 255, "1")
    return pil_squeezed_01



"""Dynamic simulator interface for LGSVL."""

import math
import time
import warnings

import lgsvl

import scenic.core.simulators as simulators
from scenic.core.vectors import Vector
import scenic.simulators.lgsvl.utils as utils
from scenic.syntax.veneer import verbosePrint


class LGSVLSimulator(simulators.Simulator):
    """A connection to an instance of LGSVL.

    See the `SVL documentation`_ for details on how to set the parameters below.

    Uses a default timestep of 0.1 seconds.

    Args:
        sceneID (str): Identifier for the map ("scene") to load in SVL.
        address (str): Address where SVL is running.
        port (int): Port on which to connect to SVL.
        alwaysReload (bool): Whether to force reloading the map upon connecting,
            even if the simulator already has the desired map loaded.

    .. _SVL documentation: https://www.svlsimulator.com/docs/python-api/python-api
    """

    def __init__(self, sceneID, address="localhost", port=8181, alwaysReload=False):
        super().__init__()
        verbosePrint("Connecting to LGSVL Simulator...")
        self.client = lgsvl.Simulator(address=address, port=port)
        if alwaysReload or self.client.current_scene != sceneID:
            self.client.load(scene=sceneID)
        verbosePrint("Map loaded in simulator.")

    def createSimulation(self, scene, **kwargs):
        return LGSVLSimulation(scene, self.client, **kwargs)


class LGSVLSimulation(simulators.Simulation):
    """Subclass of `Simulation` for LGSVL."""

    def __init__(self, scene, client, *, timestep, **kwargs):
        self.client = client
        self.usingApollo = False
        self.data = {}
        self.collisionOccurred = False

        if timestep is None:
            timestep = scene.params.get("time_step", 0.1)  # backwards-compatibility

        super().__init__(scene, timestep=timestep, **kwargs)

    def setup(self):
        # Reset simulator (deletes all existing objects)
        self.client.reset()

        super().setup()

    def createObjectInSimulator(self, obj):
        if not hasattr(obj, "lgsvlObject"):
            return  # not an LGSVL object

        # Figure out what type of LGSVL object this is
        if not hasattr(obj, "lgsvlName"):
            raise RuntimeError(f"object {obj} does not have an lgsvlName property")
        if not hasattr(obj, "lgsvlAgentType"):
            raise RuntimeError(f"object {obj} does not have an lgsvlAgentType property")
        name = obj.lgsvlName
        agentType = obj.lgsvlAgentType

        # Set up position and orientation
        state = lgsvl.AgentState()
        elevation = obj.elevation
        if elevation is None:
            elevation = self.groundElevationAt(obj.position)
        state.transform.position = utils.scenicToLGSVLPosition(obj.position, elevation)
        state.transform.rotation = utils.scenicToLGSVLRotation(obj.heading)

        # Create LGSVL object
        lgsvlObj = self.client.add_agent(name, agentType, state)
        obj.lgsvlObject = lgsvlObj

        # Initialize Data
        self.data[obj] = {}
        # Initialize Apollo if needed
        if getattr(obj, "apolloVehicle", None):
            self.initApolloFor(obj, lgsvlObj)

    def groundElevationAt(self, pos):
        origin = utils.scenicToLGSVLPosition(pos, 100000)
        result = self.client.raycast(origin, lgsvl.Vector(0, -1, 0), 1)
        if result is None:
            warnings.warn(
                f"no ground at position {pos}", simulators.SimulatorInterfaceWarning
            )
            return 0
        return result.point.y

    def initApolloFor(self, obj, lgsvlObj):
        """Initialize Apollo for an ego vehicle.

        Uses LG's interface which injects packets into Dreamview.
        """
        if self.usingApollo:
            raise RuntimeError("can only use one Apollo vehicle")
        self.usingApollo = True

        def on_collision(agent1, agent2, contact):
            if agent1 is not None and agent1.name == lgsvlObj.name:
                self.data[obj]["collision"] = True
            if agent2 is not None and agent2.name == lgsvlObj.name:
                self.data[obj]["collision"] = True
            if self.data[obj]["collision"]:
                self.collisionOccurred = True

        # Initialize Data
        self.data[obj]["collision"] = False
        lgsvlObj.on_collision(on_collision)

        # connect bridge from LGSVL to Apollo
        lgsvlObj.connect_bridge(obj.bridgeHost, obj.bridgePort)

        # set up connection and map/vehicle configuration
        from lgsvl import dreamview

        dv = dreamview.Connection(self.client, lgsvlObj)
        obj.dreamview = dv
        hdMap = self.scene.params["apolloHDMap"]
        dv.set_hd_map(hdMap)
        dv.set_vehicle(obj.apolloVehicle)

    def executeActions(self, allActions):
        super().executeActions(allActions)

        # Apply state/control updates which were accumulated while executing the actions
        for obj in self.agents:
            if obj._stateUpdated:
                obj.lgsvlObject.state = obj.state
                obj._stateUpdated = False
            ctrl = getattr(obj, "_control", None)
            if ctrl is not None:
                obj.lgsvlObject.apply_control(ctrl, obj._stickyControl)
                obj._control = None

    def step(self):
        self.client.run(time_limit=self.timestep)

    def getProperties(self, obj, properties):
        lgsvlObj = obj.lgsvlObject
        state = lgsvlObj.state
        obj.state = state  # cache state for subsequent updates

        velocity = utils.lgsvlToScenicPosition(state.velocity)
        speed = math.hypot(*velocity)

        values = dict(
            position=utils.lgsvlToScenicPosition(state.position),
            elevation=utils.lgsvlToScenicElevation(state.position),
            heading=utils.lgsvlToScenicRotation(state.rotation),
            velocity=velocity,
            speed=speed,
            angularSpeed=utils.lgsvlToScenicAngularSpeed(state.rotation),
        )
        return values

    def destroy(self):
        self.client.reset()



"""A compiler and scene generator for the Scenic scenario description language."""

import scenic.core.errors as _errors
from scenic.core.errors import setDebuggingOptions
from scenic.syntax.translator import scenarioFromFile, scenarioFromString

_errors.showInternalBacktrace = False  # see comment in errors module
del _errors



"""Scenic world model for Grand Theft Auto V (GTAV).

This interface must be used in `2D compatibility mode`.

Importing scenes generated using this model into GTA V requires a GTA V plugin,
which you can find `here <https://github.com/xyyue/scenic2gta>`__.
"""



"""Scenic world models for the Webots robotics simulator.

This module contains common code for working with Webots, e.g. parsing WBT files,
as well as a generic dynamic simulator interface and world model for Webots.
More detailed world models for particular types of scenarios are in submodules.
"""

from .simulator import WebotsSimulator
from .utils import ENU, EUN, NUE



""" Module containing the Shape class and its subclasses, which represent shapes of Objects"""

from abc import ABC, abstractmethod

import numpy
import trimesh
from trimesh.transformations import (
    concatenate_matrices,
    quaternion_matrix,
    translation_matrix,
)

from scenic.core.type_support import toOrientation
from scenic.core.utils import cached_property, unifyMesh
from scenic.core.vectors import Orientation

###################################################################################################
# Abstract Classes and Utilities
###################################################################################################


class Shape(ABC):
    """An abstract base class for Scenic shapes.

    Represents a physical shape in Scenic. Does not encode position or orientation,
    which are handled by the `Region` class. Does contain dimension information, which
    is used as a default value by any `Object` with this shape and can be overwritten.

    If dimensions and scale are both specified the dimensions are first set by dimensions,
    and then scaled by scale.

    Args:
        dimensions: The raw (before scaling) dimensions of the shape.
        scale: Scales all the dimensions of the shape by a multiplicative factor.
    """

    def __init__(self, dimensions, scale):
        # Store values
        self.raw_dimensions = dimensions
        self.scale = scale
        self.dimensions = tuple(dim * self.scale for dim in self.raw_dimensions)
        self.width = self.dimensions[0]
        self.length = self.dimensions[1]
        self.height = self.dimensions[2]
        for dim, name in zip(self.dimensions, ("width", "length", "height")):
            if dim <= 0:
                raise ValueError(f"{name} of shape must be positive")

    @cached_property
    def containsCenter(self):
        """Whether or not this object contains its central point"""
        pq = trimesh.proximity.ProximityQuery(self.mesh)
        region_distance = pq.signed_distance([(0, 0, 0)])[0]

        return region_distance > 0

    @property
    @abstractmethod
    def mesh(self):
        pass

    @property
    @abstractmethod
    def isConvex(self):
        pass


###################################################################################################
# 3D Shape Classes
###################################################################################################


class MeshShape(Shape):
    """A Shape subclass defined by a `trimesh.base.Trimesh` object.

    The mesh passed must be a `trimesh.base.Trimesh` object that represents a well defined
    volume (i.e. the ``is_volume`` property must be true), meaning the mesh must be watertight,
    have consistent winding and have outward facing normals.

    Args:
        mesh: A mesh object.
        dimensions: The raw (before scaling) dimensions of the shape. If dimensions
          and scale are both specified the dimensions are first set by dimensions, and then
          scaled by scale.
        scale: Scales all the dimensions of the shape by a multiplicative factor.
          If dimensions and scale are both specified the dimensions are first set by dimensions,
          and then scaled by scale.
        initial_rotation: A 3-tuple containing the yaw, pitch, and roll respectively to apply when loading
          the mesh. Note the initial_rotation must be fixed.
    """

    def __init__(self, mesh, dimensions=None, scale=1, initial_rotation=None):
        # Ensure the mesh is watertight so volume is well defined
        if not mesh.is_volume:
            raise ValueError(
                "A MeshShape cannot be defined with a mesh that does not have a well defined volume."
                " Consider using scenic.core.utils.repairMesh."
            )

        # Copy mesh and center vertices around origin
        self._mesh = mesh.copy()
        self._mesh.vertices -= self._mesh.bounding_box.center_mass

        # If rotation is provided, apply rotation
        if initial_rotation is not None:
            rotation = toOrientation(initial_rotation)
            rotation_matrix = quaternion_matrix(
                (rotation.w, rotation.x, rotation.y, rotation.z)
            )
            self._mesh.apply_transform(rotation_matrix)

        # If dimensions are not specified, infer them.
        if dimensions is None:
            dimensions = list(self._mesh.extents)

        # Scale mesh to unit size
        scale_vals = numpy.array(self._mesh.extents)
        scale_matrix = numpy.eye(4)
        scale_matrix[:3, :3] /= scale_vals
        self._mesh.apply_transform(scale_matrix)

        super().__init__(dimensions, scale)

    @classmethod
    def fromFile(cls, path, unify=True, **kwargs):
        """Load a mesh shape from a file, attempting to infer filetype and compression.

        For example: "foo.obj.bz2" is assumed to be a compressed .obj file.
        "foo.obj" is assumed to be an uncompressed .obj file. "foo" is an
        unknown filetype, so unless a filetype is provided an exception will be raised.

        Args:
            path (str): Path to the file to import.
            filetype (str): Filetype of file to be imported. This will be inferred if not provided.
                The filetype must be one compatible with `trimesh.load`.
            compressed (bool): Whether or not this file is compressed (with bz2). This will be inferred
                if not provided.
            binary (bool): Whether or not to open the file as a binary file.
            unify (bool): Whether or not to attempt to unify this mesh.
            kwargs: Additional arguments to the MeshShape initializer.
        """
        mesh = trimesh.load(path, force="mesh")
        if not mesh.is_volume:
            raise ValueError(
                "A MeshShape cannot be defined with a mesh that does not have a well defined volume."
                " Consider using scenic.core.utils.repairMesh."
            )
        if unify:
            mesh = unifyMesh(mesh, verbose=True)
        return cls(mesh, **kwargs)

    @property
    def mesh(self):
        return self._mesh

    @cached_property
    def isConvex(self):
        return self._mesh.is_convex

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_mesh"] = self._mesh.copy()
        return state


class BoxShape(MeshShape):
    """A box shape with all dimensions 1 by default."""

    def __init__(self, dimensions=(1, 1, 1), scale=1, initial_rotation=None):
        super().__init__(
            trimesh.creation.box((1, 1, 1)), dimensions, scale, initial_rotation
        )


class CylinderShape(MeshShape):
    """A cylinder shape with all dimensions 1 by default."""

    def __init__(self, dimensions=(1, 1, 1), scale=1, initial_rotation=None, sections=24):
        super().__init__(
            trimesh.creation.cylinder(radius=0.5, height=1, sections=sections),
            dimensions,
            scale,
            initial_rotation,
        )
        self.sections = sections


class ConeShape(MeshShape):
    """A cone shape with all dimensions 1 by default."""

    def __init__(self, dimensions=(1, 1, 1), scale=1, initial_rotation=None):
        super().__init__(
            trimesh.creation.cone(radius=0.5, height=1),
            dimensions,
            scale,
            initial_rotation,
        )


class SpheroidShape(MeshShape):
    """A spheroid shape with all dimensions 1 by default."""

    def __init__(self, dimensions=(1, 1, 1), scale=1, initial_rotation=None):
        super().__init__(
            trimesh.creation.icosphere(radius=1), dimensions, scale, initial_rotation
        )



"""Abstract interface to simulators supporting the driving domain."""

from scenic.core.simulators import Simulation, Simulator
from scenic.domains.driving.controllers import (
    PIDLateralController,
    PIDLongitudinalController,
)


class DrivingSimulator(Simulator):
    """A `Simulator` supporting the driving domain."""

    pass


class DrivingSimulation(Simulation):
    """A `Simulation` with a simulator supporting the driving domain.

    This subclass of `Simulation` provides no special behavior by itself; it
    just provides convenience methods for creating controllers to be used by
    `FollowLaneBehavior` and related behaviors, so that the parameters of these
    controllers can be customized for different simulators.
    """

    def getLaneFollowingControllers(self, agent):
        """Get longitudinal and lateral controllers for lane following.

        The default controllers are simple PID controllers with parameters that
        work reasonably well for cars in simulators with realistic physics. See the
        classes `PIDLongitudinalController` and `PIDLateralController` for details,
        and `NewtonianSimulation` for an example of how to override this function.

        Returns:
            A pair of controllers for throttle and steering respectively.
        """
        dt = self.timestep
        if agent.isCar:
            lon_controller = PIDLongitudinalController(K_P=0.5, K_D=0.1, K_I=0.7, dt=dt)
            lat_controller = PIDLateralController(K_P=0.2, K_D=0.1, K_I=0.0, dt=dt)
        else:
            lon_controller = PIDLongitudinalController(
                K_P=0.25, K_D=0.025, K_I=0.0, dt=dt
            )
            lat_controller = PIDLateralController(K_P=0.2, K_D=0.1, K_I=0.0, dt=dt)
        return lon_controller, lat_controller

    def getTurningControllers(self, agent):
        """Get longitudinal and lateral controllers for turning."""
        dt = self.timestep
        if agent.isCar:
            lon_controller = PIDLongitudinalController(K_P=0.5, K_D=0.1, K_I=0.7, dt=dt)
            lat_controller = PIDLateralController(K_P=0.8, K_D=0.2, K_I=0.0, dt=dt)
        else:
            lon_controller = PIDLongitudinalController(
                K_P=0.25, K_D=0.025, K_I=0.0, dt=dt
            )
            lat_controller = PIDLateralController(K_P=0.4, K_D=0.1, K_I=0.0, dt=dt)
        return lon_controller, lat_controller

    def getLaneChangingControllers(self, agent):
        """Get longitudinal and lateral controllers for lane changing."""
        dt = self.timestep
        if agent.isCar:
            lon_controller = PIDLongitudinalController(K_P=0.5, K_D=0.1, K_I=0.7, dt=dt)
            lat_controller = PIDLateralController(K_P=0.08, K_D=0.3, K_I=0.0, dt=dt)
        else:
            lon_controller = PIDLongitudinalController(
                K_P=0.25, K_D=0.025, K_I=0.0, dt=dt
            )
            lat_controller = PIDLateralController(K_P=0.1, K_D=0.3, K_I=0.0, dt=dt)
        return lon_controller, lat_controller



"""Utility functions for geometric computation."""

import itertools
import math
import warnings

import numpy as np
import shapely.geometry
import shapely.ops

from scenic.core.distributions import (
    distributionFunction,
    monotonicDistributionFunction,
    needsSampling,
)
from scenic.core.lazy_eval import isLazy
from scenic.core.utils import cached_property


@distributionFunction
def sin(x) -> float:
    return math.sin(x)


@distributionFunction
def cos(x) -> float:
    return math.cos(x)


@monotonicDistributionFunction
def hypot(*args) -> float:
    return math.hypot(*args)


@monotonicDistributionFunction
def max(*args, **kwargs):
    return __builtins__["max"](*args, **kwargs)


@monotonicDistributionFunction
def min(*args, **kwargs):
    return __builtins__["min"](*args, **kwargs)


@distributionFunction
def normalizeAngle(angle) -> float:
    while angle > math.pi:
        angle -= math.tau
    while angle < -math.pi:
        angle += math.tau
    assert -math.pi <= angle <= math.pi
    return angle


def averageVectors(a, b, weight=0.5):
    ax, ay = a[0], a[1]
    bx, by = b[0], b[1]
    aw, bw = 1.0 - weight, weight
    return (ax * aw + bx * bw, ay * aw + by * bw)


def rotateVector(vector, angle):
    x, y = vector
    c, s = cos(angle), sin(angle)
    return ((c * x) - (s * y), (s * x) + (c * y))


def findMinMax(iterable):
    minv = float("inf")
    maxv = float("-inf")
    for val in iterable:
        if val < minv:
            minv = val
        if val > maxv:
            maxv = val
    return (minv, maxv)


def headingOfSegment(pointA, pointB):
    ax, ay = pointA[:2]
    bx, by = pointB[:2]
    return normalizeAngle(math.atan2(by - ay, bx - ax) - (math.pi / 2.0))


def viewAngleToPoint(point, base, heading):
    x, y, _ = base
    ox, oy, _ = point
    a = math.atan2(oy - y, ox - x) - (heading + (math.pi / 2.0))
    return normalizeAngle(a)


def apparentHeadingAtPoint(point, heading, base):
    x, y = base[:2]
    ox, oy = point[:2]
    a = (heading + (math.pi / 2.0)) - math.atan2(oy - y, ox - x)
    return normalizeAngle(a)


def pointIsInCone(point, base, heading, angle):
    va = viewAngleToPoint(point, base, heading)
    return abs(va) <= angle / 2.0


def distanceToLine(point, a, b):
    lx, ly = b[0] - a[0], b[1] - a[1]
    norm = math.hypot(lx, ly)
    nx, ny = -ly / norm, lx / norm
    px, py = point[0] - a[0], point[1] - a[1]
    return abs((px * nx) + (py * ny))


# Fastest known way to make a Shapely Point from a list/tuple/Vector
makeShapelyPoint = shapely.lib.points


def polygonUnion(polys, buf=0, tolerance=0, holeTolerance=0.002):
    if not polys:
        return shapely.geometry.Polygon()
    polys = list(polys)
    if len(polys) == 1:
        return polys[0]
    buffered = [poly.buffer(buf) for poly in polys]
    # remove empty polys to avoid triggering segfault in GEOS 3.10
    # (see https://github.com/Toblerity/Shapely/issues/1230)
    nonempty = [poly for poly in buffered if not poly.is_empty]
    union = shapely.ops.unary_union(nonempty).buffer(-buf)
    assert union.is_valid, union
    if tolerance > 0:
        union = cleanPolygon(union, tolerance, holeTolerance)
        assert union.is_valid, union
    return union


def cleanPolygon(poly, tolerance, holeTolerance=0, minRelArea=0.05, minHullLenRatio=0.9):
    if holeTolerance == 0:
        holeTolerance = tolerance * tolerance
    if poly.is_empty:
        return poly
    elif isinstance(poly, shapely.geometry.MultiPolygon):
        polys = [cleanPolygon(p, tolerance, holeTolerance) for p in poly.geoms]
        total = sum(poly.area for poly in polys)
        kept = []
        for poly in polys:
            area = poly.area
            if area >= holeTolerance or area >= minRelArea * total:
                kept.append(poly)
        poly = shapely.ops.unary_union(kept)
        assert poly.is_valid, poly
        return poly
    exterior = poly.exterior.simplify(tolerance)
    newExterior = cleanChain(exterior.coords, tolerance)
    if len(newExterior) <= 3:
        # attempt to save very thin polygons that would get reduced to a single point
        hull = exterior.convex_hull
        if hull.length >= minHullLenRatio * exterior.length:
            return hull
        return shapely.geometry.Polygon()
    ints = []
    for interior in poly.interiors:
        interior = interior.simplify(tolerance)
        interior = cleanChain(interior.coords, tolerance)
        if len(interior) >= 4:
            hole = shapely.geometry.Polygon(interior)
            if hole.area > holeTolerance:
                ints.append(interior)
    newPoly = shapely.geometry.Polygon(newExterior, ints)
    if not newPoly.is_valid:
        # last-ditch attempt to salvage polygon by splitting across self-intersections
        ext = splitSelfIntersections(
            newPoly.exterior,
            minArea=holeTolerance,
            minRelArea=minRelArea,
            minHullLenRatio=minHullLenRatio,
        )
        holes = [
            splitSelfIntersections(
                hole,
                minArea=holeTolerance,
                minRelArea=minRelArea,
                minHullLenRatio=minHullLenRatio,
            )
            for hole in newPoly.interiors
        ]
        newPoly = ext.difference(shapely.ops.unary_union(holes))
    assert newPoly.is_valid, newPoly
    return newPoly


def splitSelfIntersections(chain, minArea, minRelArea, minHullLenRatio):
    ls = shapely.geometry.LineString(chain)
    parts = list(shapely.ops.polygonize(shapely.ops.unary_union(ls)))
    total = sum(part.area for part in parts)
    if total < minArea:
        # attempt to save very thin polygons that could get thrown out by polygonize
        hull = ls.convex_hull
        if hull.length >= minHullLenRatio * ls.length:
            return hull
    kept = [
        part
        for part in parts
        if part.area >= minArea or (part.area / total) >= minRelArea
    ]
    return shapely.ops.unary_union(kept)


def cleanChain(chain, tolerance=1e-6, lineTolerance=1e-6):
    closed = tuple(chain[0]) == tuple(chain[-1])
    minLength = 4 if closed else 3
    if len(chain) <= minLength:
        return chain
    tol2 = tolerance * tolerance
    # collapse nearby points (since Shapely's simplify method doesn't always do this)
    a = chain[0]
    newChain = [a]
    for b in chain[1:]:
        dx, dy = b[0] - a[0], b[1] - a[1]
        if (dx * dx) + (dy * dy) > tol2:
            newChain.append(b)
            a = b
    if closed:
        b = chain[0]
        dx, dy = b[0] - a[0], b[1] - a[1]
        if (dx * dx) + (dy * dy) <= tol2:
            newChain.pop()
        newChain.append(chain[0])
    if len(newChain) <= minLength:
        if not closed:
            if len(newChain) > 1:
                newChain.pop()
            newChain.append(chain[-1])
        return newChain
    # collapse collinear points
    chain = newChain
    if closed:
        a = chain[-2]
        b = chain[0]
        ci = 1
        newChain = []
    else:
        a = chain[0]
        b = chain[1]
        ci = 2
        newChain = [a]
    for c in chain[ci:]:
        dx, dy = c[0] - a[0], c[1] - a[1]
        if dx == dy == 0 or distanceToLine(b, a, c) > lineTolerance:
            newChain.append(b)
            a = b
            b = c
        else:
            b = c
    newChain.append(c)
    return newChain


def removeHoles(polygon):
    if polygon.is_empty:
        return polygon
    elif isinstance(polygon, shapely.geometry.MultiPolygon):
        polys = (removeHoles(poly) for poly in polygon.geoms)
        poly = shapely.geometry.MultiPolygon(polys)
        assert poly.is_valid, poly
        return poly
    else:
        return shapely.geometry.Polygon(polygon.exterior)


class TriangulationError(RuntimeError):
    """Signals that the installed triangulation libraries are insufficient."""

    pass


def triangulatePolygon(polygon):
    """Triangulate the given Shapely polygon.

    Note that we can't use ``shapely.ops.triangulate`` since it triangulates
    point sets, not polygons (i.e., it doesn't respect edges). We need an
    algorithm for triangulation of polygons with holes (it doesn't need to be a
    Delaunay triangulation).

    Args:
        polygon (shapely.geometry.Polygon): Polygon to triangulate.

    Returns:
        A list of disjoint (except for edges) triangles whose union is the
        original polygon.
    """
    try:
        return triangulatePolygon_mapbox(polygon)
    except ImportError:
        pass
    raise RuntimeError(
        "no triangulation libraries installed " "(did you uninstall mapbox_earcut?)"
    )


def triangulatePolygon_mapbox(polygon):
    import mapbox_earcut

    vertices, rings = [], []
    ring = polygon.exterior.coords[:-1]  # despite 'ring' name, actually need a chain
    vertices.extend(ring)
    rings.append(len(vertices))
    for interior in polygon.interiors:
        ring = interior.coords[:-1]
        vertices.extend(ring)
        rings.append(len(vertices))
    vertices = np.array(vertices, dtype=np.float64)[:, :2]
    rings = np.array(rings)
    result = mapbox_earcut.triangulate_float64(vertices, rings)

    triangles = []
    points = vertices[result]
    coords = np.split(points, len(points) / 3)
    triangles = shapely.polygons(coords)
    return triangles


def allChains(polygon):
    if polygon.is_empty:
        return
    if isinstance(
        polygon,
        (
            shapely.geometry.MultiPolygon,
            shapely.geometry.MultiLineString,
            shapely.geometry.MultiPoint,
            shapely.geometry.collection.GeometryCollection,
        ),
    ):
        polygons = polygon.geoms
    else:
        polygons = [polygon]
    for polygon in polygons:
        if isinstance(polygon, shapely.geometry.Polygon):
            yield polygon.exterior
            for ring in polygon.interiors:
                yield ring
        elif isinstance(
            polygon,
            (
                shapely.geometry.LineString,
                shapely.geometry.LinearRing,
                shapely.geometry.Point,
            ),
        ):
            yield polygon
        else:
            raise RuntimeError(f"unknown kind of shapely geometry {polygon}")


def plotPolygon(polygon, plt, style="r-", **kwargs):
    for chain in allChains(polygon):
        x, y = chain.xy
        plt.plot(x, y, style, **kwargs)



"""Interface to the CARLA driving simulator.

This interface must currently be used in `2D compatibility mode`.

This interface has been tested with `CARLA <https://carla.org/>`_ versions 0.9.9,
0.9.10, and 0.9.11.
It supports dynamic scenarios involving vehicles, pedestrians, and props.

The interface implements the :obj:`scenic.domains.driving` abstract domain, so any
object types, behaviors, utility functions, etc. from that domain may be used freely.
For details of additional CARLA-specific functionality, see the world model
:obj:`scenic.simulators.carla.model`.
"""

# Only import CarlaSimulator if the carla package is installed; otherwise the
# import would raise an exception.
carla = None
try:
    import carla
except ImportError:
    pass
if carla:
    from .simulator import CarlaSimulator
del carla



"""Assorted utilities and classes used throughout the dynamics package."""


class RejectSimulationException(Exception):
    """Exception indicating a requirement was violated at runtime."""

    pass


class StuckBehaviorWarning(UserWarning):
    """Warning issued when a behavior/scenario may have gotten stuck.

    When a behavior or compose block of a modular scenario executes for a long
    time without yielding control, there is no way to tell whether it has
    entered an infinite loop with no take/wait statements, or is actually doing
    some long computation. But since forgetting a wait statement in a while loop
    is an easy mistake, we raise this warning after a behavior/scenario has run
    for `stuckBehaviorWarningTimeout` seconds without yielding.
    """

    pass



"""Simple Newtonian physics simulator.

This interface must be used in `2D compatibility mode`.

This simulator allows dynamic scenarios to be tested without installing an
external simulator. It is currently very simplistic (e.g. not modeling
collisions).

The simulator provides two world models: a generic one, and a more specialized
model supporting traffic scenarios using the :obj:`scenic.domains.driving`
abstract domain.
"""

from .simulator import NewtonianSimulator



"""Specifiers and associated objects."""

import itertools

from scenic.core.distributions import toDistribution
from scenic.core.errors import InvalidScenarioError, SpecifierError
from scenic.core.lazy_eval import (
    DelayedArgument,
    needsLazyEvaluation,
    requiredProperties,
    toLazyValue,
    valueInContext,
)

## Specifiers themselves


class Specifier:
    """Specifier providing values for properties.

    Each property is set to a value, at a given priority,
    given dependencies.

    Args:
        name: The name of this specifier.
        priorities: A dictionary mapping properties to the priority
          they are being specified with.
        value: A dictionary mapping properties to the values they are
          being specified as.
        deps: An iterable containing all properties that this
          specifier relies on.
    """

    def __init__(self, name, priorities, value, deps=None):
        assert isinstance(priorities, dict)
        assert isinstance(value, (dict, DelayedArgument))

        self.priorities = priorities
        self.value = toLazyValue(value)

        if deps is None:
            deps = set()

        deps |= requiredProperties(self.value)

        for p in priorities:
            if p in deps:
                raise SpecifierError(f"specifier for property {p} depends on itself")

        self.requiredProperties = tuple(sorted(deps))
        self.name = name

    def getValuesFor(self, obj):
        """Get the values specified for a given object."""
        val = valueInContext(self.value, obj)
        assert isinstance(val, dict)
        return val

    def __str__(self):
        return f"<{self.name} Specifier for {self.priorities}>"


class ModifyingSpecifier(Specifier):
    """Specifier providing values (or modifying) properties.

    Args:
        name: The name of this specifier.
        priorities: A dictionary mapping properties to the priority
          they are being specified with.
        value: A dictionary mapping properties to the values they are
          being specified as.
        modifiable_props: What properties specified by this specifier
          can be modified.
        deps: An iterable containing all properties that this
          specifier relies on.
    """

    def __init__(self, name, priorities, value, modifiable_props, deps=None):
        self.modifiable_props = modifiable_props
        super().__init__(name, priorities, value, deps)


## Support for property defaults


class PropertyDefault:
    """A default value, possibly with dependencies."""

    def __init__(self, requiredProperties, attributes, value):
        self.requiredProperties = set(requiredProperties)
        self.value = value

        def enabled(thing, default):
            if thing in attributes:
                attributes.remove(thing)
                return True
            else:
                return default

        self.isAdditive = enabled("additive", False)
        self.isDynamic = enabled("dynamic", False)
        self.isFinal = enabled("final", False)
        for attr in attributes:
            raise RuntimeError(f'unknown property attribute "{attr}"')
        if self.isAdditive and self.isDynamic:
            raise InvalidScenarioError("additive properties cannot be dynamic")

    @staticmethod
    def forValue(value):
        if isinstance(value, PropertyDefault):
            return value
        else:
            return PropertyDefault(set(), set(), lambda self: value)

    def resolveFor(self, prop, overriddenDefs):
        """Create a Specifier for a property from this default and any superclass defaults."""
        for other in overriddenDefs:
            if other.isFinal:
                msg = f'"{prop}" property cannot be overridden'
                if prop == "heading":
                    msg += " (perhaps this scenario requires the --2d flag?)"
                raise InvalidScenarioError(msg)
        if self.isAdditive:
            allReqs = self.requiredProperties
            for other in overriddenDefs:
                allReqs |= other.requiredProperties

            def concatenator(context):
                allVals = [self.value(context)]
                for other in overriddenDefs:
                    allVals.append(other.value(context))
                return tuple(allVals)

            val = DelayedArgument(allReqs, concatenator, _internal=True)
        else:
            val = DelayedArgument(self.requiredProperties, self.value, _internal=True)

        return Specifier("PropertyDefault", {prop: -1}, {prop: val})



"""Low-level controllers useful for vehicles.

The Lateral/Longitudinal PID controllers are adapted from `CARLA`_'s PID controllers,
which are licensed under the following terms:

    Copyright (c) 2018-2020 CVC.

    This work is licensed under the terms of the MIT license.
    For a copy, see <https://opensource.org/licenses/MIT>.


.. _CARLA: https://carla.org/
"""

from collections import deque

import numpy as np


class PIDLongitudinalController:
    """Longitudinal control using a PID to reach a target speed.

    Arguments:
        K_P: Proportional gain
        K_D: Derivative gain
        K_I: Integral gain
        dt: time step
    """

    def __init__(self, K_P=0.5, K_D=0.1, K_I=0.2, dt=0.1):
        self._k_p = K_P
        self._k_d = K_D
        self._k_i = K_I
        self._dt = dt
        self._error_buffer = deque(maxlen=10)

    def run_step(self, speed_error):
        """Estimate the throttle/brake of the vehicle based on the PID equations.

        Arguments:
            speed_error: target speed minus current speed

        Returns:
            a signal between -1 and 1, with negative values indicating braking.
        """
        error = speed_error
        self._error_buffer.append(error)

        if len(self._error_buffer) >= 2:
            _de = (self._error_buffer[-1] - self._error_buffer[-2]) / self._dt
            _ie = sum(self._error_buffer) * self._dt
        else:
            _de = 0.0
            _ie = 0.0

        return np.clip(
            (self._k_p * error) + (self._k_d * _de) + (self._k_i * _ie), -1.0, 1.0
        )


class PIDLateralController:
    """Lateral control using a PID to track a trajectory.

    Arguments:
        K_P: Proportional gain
        K_D: Derivative gain
        K_I: Integral gain
        dt: time step
    """

    def __init__(self, K_P=0.3, K_D=0.2, K_I=0, dt=0.1):
        self.Kp = K_P
        self.Kd = K_D
        self.Ki = K_I
        self.PTerm = 0
        self.ITerm = 0
        self.DTerm = 0
        self.dt = dt
        self.last_error = 0
        self.windup_guard = 20.0
        self.output = 0

    def run_step(self, cte):
        """Estimate the steering angle of the vehicle based on the PID equations.

        Arguments:
            cte: cross-track error (distance to right of desired trajectory)

        Returns:
            a signal between -1 and 1, with -1 meaning maximum steering to the left.
        """
        error = cte
        delta_error = error - self.last_error
        self.PTerm = self.Kp * error
        self.ITerm += error * self.dt

        if self.ITerm < -self.windup_guard:
            self.ITerm = -self.windup_guard
        elif self.ITerm > self.windup_guard:
            self.ITerm = self.windup_guard

        self.DTerm = delta_error / self.dt

        # Remember last error for next calculation
        self.last_error = error

        self.output = self.PTerm + (self.Ki * self.ITerm) + (self.Kd * self.DTerm)

        return np.clip(self.output, -1, 1)



# This file utilizes existing CARLA code:

# Copyright (c) 2018 Intel Labs.
# authors: German Ros (german.ros@intel.com)
#
# This work is licensed under the terms of the MIT license.
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import math
import weakref

import carla
from carla import ColorConverter as cc
import numpy as np
import pygame


def get_actor_display_name(actor, truncate=250):
    name = " ".join(actor.type_id.replace("_", ".").title().split(".")[1:])
    return (name[: truncate - 1] + "\u2026") if len(name) > truncate else name


# ==============================================================================
# -- HUD -----------------------------------------------------------------------
# ==============================================================================


class HUD(object):
    def __init__(self, width, height):
        self.dim = (width, height)
        font = pygame.font.Font(pygame.font.get_default_font(), 20)
        fonts = [x for x in pygame.font.get_fonts() if "mono" in x]
        default_font = "ubuntumono"
        mono = default_font if default_font in fonts else fonts[0]
        mono = pygame.font.match_font(mono)
        self._font_mono = pygame.font.Font(mono, 14)
        self._notifications = FadingText(font, (width, 40), (0, height - 40))
        self.server_fps = 0
        self.frame = 0
        self.simulation_time = 0
        self._info_text = []
        self._server_clock = pygame.time.Clock()

    def on_world_tick(self, timestamp):
        self._server_clock.tick()
        self.server_fps = self._server_clock.get_fps()
        self.frame = timestamp.frame
        self.simulation_time = timestamp.elapsed_seconds

    def tick(self, world, ego, clock, showLabels=True):
        if ego.carlaActor is None:
            return  # ego not spawned yet

        t = ego.carlaActor.get_transform()
        v = ego.carlaActor.get_velocity()
        c = ego.carlaActor.get_control()

        heading = "N" if abs(t.rotation.yaw) < 89.5 else ""
        heading += "S" if abs(t.rotation.yaw) > 90.5 else ""
        heading += "E" if 179.5 > t.rotation.yaw > 0.5 else ""
        heading += "W" if -0.5 > t.rotation.yaw > -179.5 else ""

        # colhist = collisionSensor.get_collision_history()
        # collision = [colhist[x + self.frame - 200] for x in range(0, 200)]
        # max_col = max(1.0, max(collision))
        # collision = [x / max_col for x in collision]

        vehicles = world.get_actors().filter("vehicle.*")
        pedestrians = world.get_actors().filter("walker.pedestrian.*")

        self._info_text = [
            "Server:  % 16d FPS" % self.server_fps,
            "Map:    % 20s" % world.get_map().name,
            "Simulation time: % 12s"
            % datetime.timedelta(seconds=int(self.simulation_time)),
            "",
            "Number of vehicles: % 8d" % len(vehicles),
            "Number of pedestrians: % 8d" % len(pedestrians),
            "",
            "Ego: % 20s" % get_actor_display_name(ego.carlaActor, truncate=20),
            "",
            "Speed:   % 15.0f m/s" % math.sqrt(v.x**2 + v.y**2 + v.z**2),
            "Heading:% 16.0f\N{DEGREE SIGN} % 2s" % (t.rotation.yaw, heading),
            "Location:% 20s" % ("(% 5.3f, % 5.3f)" % (t.location.x, t.location.y)),
            "Height:  % 18.0f m" % t.location.z,
        ]

        try:
            _control_text = [
                "",
                ("Throttle:", c.throttle, 0.0, 1.0),
                ("Steer:", c.steer, -1.0, 1.0),
                ("Brake:", c.brake, 0.0, 1.0),
                ("Reverse:", c.reverse),
                ("Hand brake:", c.hand_brake),
                ("Manual:", c.manual_gear_shift),
                "Gear:      %s" % {-1: "R", 0: "N"}.get(c.gear, c.gear),
                #'',
                #'Collision:', collision,
            ]
        except:
            _control_text = []
        finally:
            self._info_text.extend(_control_text)

        if len(vehicles) > 1:
            self._info_text += ["Nearby vehicles:"]
            distance = lambda l: math.sqrt(
                (l.x - t.location.x) ** 2
                + (l.y - t.location.y) ** 2
                + (l.z - t.location.z) ** 2
            )
            vehicles = [
                (distance(x.get_location()), x)
                for x in vehicles
                if x.id != ego.carlaActor.id
            ]
            for d, vehicle in sorted(vehicles):
                if d > 200.0:
                    break
                vehicle_type = get_actor_display_name(vehicle, truncate=22)
                self._info_text.append("% 4dm %s" % (d, vehicle_type))

    def render(self, display):
        info_surface = pygame.Surface((220, self.dim[1]))
        info_surface.set_alpha(100)
        display.blit(info_surface, (0, 0))
        v_offset = 4
        bar_h_offset = 100
        bar_width = 106
        for item in self._info_text:
            if v_offset + 18 > self.dim[1]:
                break
            if isinstance(item, list):
                if len(item) > 1:
                    points = [
                        (x + 8, v_offset + 8 + (1.0 - y) * 30) for x, y in enumerate(item)
                    ]
                    pygame.draw.lines(display, (255, 136, 0), False, points, 2)
                item = None
                v_offset += 18
            elif isinstance(item, tuple):
                if isinstance(item[1], bool):
                    rect = pygame.Rect((bar_h_offset, v_offset + 8), (6, 6))
                    pygame.draw.rect(display, (255, 255, 255), rect, 0 if item[1] else 1)
                else:
                    rect_border = pygame.Rect(
                        (bar_h_offset, v_offset + 8), (bar_width, 6)
                    )
                    pygame.draw.rect(display, (255, 255, 255), rect_border, 1)
                    f = (item[1] - item[2]) / (item[3] - item[2])
                    if item[2] < 0.0:
                        rect = pygame.Rect(
                            (bar_h_offset + f * (bar_width - 6), v_offset + 8), (6, 6)
                        )
                    else:
                        rect = pygame.Rect(
                            (bar_h_offset, v_offset + 8), (f * bar_width, 6)
                        )
                    pygame.draw.rect(display, (255, 255, 255), rect)
                item = item[0]
            if item:  # At this point has to be a str
                surface = self._font_mono.render(item, True, (255, 255, 255))
                display.blit(surface, (8, v_offset))
            v_offset += 18


# ==============================================================================
# -- FadingText ----------------------------------------------------------------
# ==============================================================================


class FadingText(object):
    def __init__(self, font, dim, pos):
        self.font = font
        self.dim = dim
        self.pos = pos
        self.seconds_left = 0
        self.surface = pygame.Surface(self.dim)

    def set_text(self, text, color=(255, 255, 255), seconds=2.0):
        text_texture = self.font.render(text, True, color)
        self.surface = pygame.Surface(self.dim)
        self.seconds_left = seconds
        self.surface.fill((0, 0, 0, 0))
        self.surface.blit(text_texture, (10, 11))

    def tick(self, _, clock):
        delta_seconds = 1e-3 * clock.get_time()
        self.seconds_left = max(0.0, self.seconds_left - delta_seconds)
        self.surface.set_alpha(500.0 * self.seconds_left)

    def render(self, display):
        display.blit(self.surface, self.pos)


# ==============================================================================
# -- CollisionSensor -----------------------------------------------------------
# ==============================================================================


class CollisionSensor(object):
    def __init__(self, world, actor, hud=None):
        self.sensor = None
        self._history = []
        self._actor = actor
        self._actor_mass = actor.get_physics_control().mass
        self._hud = hud
        self._world = world
        bp = self._world.get_blueprint_library().find("sensor.other.collision")
        self.sensor = self._world.spawn_actor(
            bp, carla.Transform(), attach_to=self._actor
        )
        # Pass the lambda a weak reference to self to avoid circular reference
        weak_self = weakref.ref(self)
        self.sensor.listen(lambda event: CollisionSensor._on_collision(weak_self, event))

    def get_collision_speeds(self):
        """Convert collision intensities from momentem (kg*m/s) to speed (m/s)."""
        return [(c[0], c[1] / self._actor_mass) for c in self._history]

    def get_collision_history(self):
        history = collections.defaultdict(int)
        for frame, intensity in self._history:
            history[frame] += intensity
        return history

    @staticmethod
    def _on_collision(weak_self, event):
        self = weak_self()
        if not self:
            return
        actor_type = get_actor_display_name(event.other_actor)
        if self._hud:
            self._hud.notification(
                "Collision with %r, id = %d" % (actor_type, event.other_actor.id)
            )
        impulse = event.normal_impulse
        intensity = math.sqrt(impulse.x**2 + impulse.y**2 + impulse.z**2)
        self._history.append((event.frame, intensity))
        if len(self._history) > 4000:
            self._history.pop(0)


# ==============================================================================
# -- CameraManager -------------------------------------------------------------
# ==============================================================================


class CameraManager(object):
    def __init__(self, world, actor, hud):
        self.sensor = None
        self._surface = None
        self._actor = actor
        self._hud = hud
        self.images = []
        self._camera_transforms = [
            carla.Transform(carla.Location(x=-5.5, z=2.8), carla.Rotation(pitch=-15)),
            carla.Transform(carla.Location(x=1.6, z=1.7)),
        ]
        self._transform_index = 1
        self._sensors = [
            ["sensor.camera.rgb", cc.Raw, "Camera RGB"],
            ["sensor.camera.depth", cc.Raw, "Camera Depth (Raw)"],
            ["sensor.camera.depth", cc.Depth, "Camera Depth (Gray Scale)"],
            [
                "sensor.camera.depth",
                cc.LogarithmicDepth,
                "Camera Depth (Logarithmic Gray Scale)",
            ],
            [
                "sensor.camera.semantic_segmentation",
                cc.Raw,
                "Camera Semantic Segmentation (Raw)",
            ],
            [
                "sensor.camera.semantic_segmentation",
                cc.CityScapesPalette,
                "Camera Semantic Segmentation (CityScapes Palette)",
            ],
            ["sensor.lidar.ray_cast", None, "Lidar (Ray-Cast)"],
        ]
        self._world = world
        bp_library = self._world.get_blueprint_library()
        for item in self._sensors:
            bp = bp_library.find(item[0])
            if item[0].startswith("sensor.camera"):
                bp.set_attribute("image_size_x", str(hud.dim[0]))
                bp.set_attribute("image_size_y", str(hud.dim[1]))
            item.append(bp)
        self._index = None

    def toggle_camera(self):
        set_transform((self._transform_index + 1) % len(self._camera_transforms))

    def set_transform(self, idx):
        self._transform_index = idx
        self.sensor.set_transform(self._camera_transforms[self._transform_index])

    def set_sensor(self, index):
        index = index % len(self._sensors)
        needs_respawn = (
            True
            if self._index is None
            else self._sensors[index][0] != self._sensors[self._index][0]
        )
        if needs_respawn:
            if self.sensor is not None:
                self.sensor.destroy()
                self._surface = None
            self.sensor = self._world.spawn_actor(
                self._sensors[index][-1],
                self._camera_transforms[self._transform_index],
                attach_to=self._actor,
            )
            # Pass lambda a weak reference to self to avoid circular reference
            weak_self = weakref.ref(self)
            self.sensor.listen(lambda image: CameraManager._parse_image(weak_self, image))
        self._index = index

    def render(self, display):
        if self._surface is not None:
            display.blit(self._surface, (0, 0))

    @staticmethod
    def _parse_image(weak_self, image):
        self = weak_self()
        if not self:
            return
        if self._sensors[self._index][0].startswith("sensor.lidar"):
            points = np.frombuffer(image.raw_data, dtype=np.dtype("f4"))
            points = np.reshape(points, (int(points.shape[0] / 3), 3))
            lidar_data = np.array(points[:, :2])
            lidar_data *= min(self._hud.dim) / 100.0
            lidar_data += (0.5 * self._hud.dim[0], 0.5 * self._hud.dim[1])
            lidar_data = np.fabs(lidar_data)
            lidar_data = lidar_data.astype(np.int32)
            lidar_data = np.reshape(lidar_data, (-1, 2))
            lidar_img_size = (self._hud.dim[0], self._hud.dim[1], 3)
            lidar_img = np.zeros(lidar_img_size)
            lidar_img[tuple(lidar_data.T)] = (255, 255, 255)
            self._surface = pygame.surfarray.make_surface(lidar_img)
        else:
            image.convert(self._sensors[self._index][1])
            array = np.frombuffer(image.raw_data, dtype=np.dtype("uint8"))
            array = np.reshape(array, (image.height, image.width, 4))
            array = array[:, :, :3]
            array = array[:, :, ::-1]
            self._surface = pygame.surfarray.make_surface(array.swapaxes(0, 1))
        self.images.append(image)

    def destroy_sensor(self):
        if self.sensor is not None:
            self.sensor.stop()
            self.sensor.destroy()



"""Support for dynamic behaviors and modular scenarios.

A few classes are exposed here for external use, including:

* `Action`;
* `GuardViolation`, `InvariantViolation`, and `PreconditionViolation`;
* `StuckBehaviorWarning`.

Everything else defined in the submodules is an implementation detail and
should not be used outside of Scenic (it may change at any time).
"""

from .actions import Action
from .guards import GuardViolation, InvariantViolation, PreconditionViolation
from .utils import RejectSimulationException, StuckBehaviorWarning

#: Timeout in seconds after which a `StuckBehaviorWarning` will be raised.
stuckBehaviorWarningTimeout = 10
